from typing import Union, Optional, Callable

from gesund.core.schema import UserInputParams, UserInputData
from gesund.core._managers.plot_manager import plot_manager


class ValidationResult:
    def __init__(
        self,
        data: UserInputData,
        input_params: UserInputParams,
        result: Union[dict, list],
        cohort_args: Optional[dict] = {},
        plot_args: Optional[dict] = {},
    ) -> None:
        """
        A function to initialize the resultant data

        :param data: the data loaded by the validation class
        :type data: UserInputData
        :param input_params: the user input provided by the user inputs
        :type input_params: UserInputParams
        :param result: result from validation execution
        :type result: Union[dict, list]
        :param cohort_args: user arguments for cohort creation
        :type cohort_args: dict
        :param  plot_args: user arguments for plotting
        :type plot_args: dict

        :return: None
        """
        self.data = data
        self.user_params = input_params
        self.result = result
        self.cohort_args = cohort_args
        self.plot_args = plot_args

    def save(self, metric_name: str = "all", format: str = "json") -> str:
        """
        A function to save the metric in json format

        :param metric_name: name of the metric to save the results
        :type metric_name: str
        :param format: data format for the result to save in
        :type format: str

        :return: location of the json file stored
        :rtype: str
        """
        pass

    def _run_plot_executor(
        self,
        plot_executor: Callable,
        metric_name: str,
        save_plot: bool,
        cohort_id: Optional[int] = None,
        file_name: Optional[str] = None,
    ) -> None:
        """
        A function to run the plot executor

        :param plot_executor: The registered function required for executing the function
        :type plot_executor: Callable
        :param metric_name: name of the metric to be plotted
        :type metric_name: str
        :param save_plot: boolean value of the metric to be saved
        :type save_plot: bool
        :param cohort_id: the id of the cohort if applicable
        :type cohort_id: int
        :param file_name: the name of the file to use for saving
        :type file_name: str

        """
        # cohort wise plotting is only possible when the cohort map is populated
        # if it is not populated then it is considered as a whole data
        if self.cohort_args["cohort_map"]:
            if cohort_id:
                result = self.result[cohort_id][metric_name]
                plot_executor(
                    results=result,
                    save_plot=save_plot,
                    cohort_id=cohort_id,
                    file_name=file_name,
                )
            else:
                for _cohort_id in self.cohort_args["cohort_map"]:
                    result = self.result[_cohort_id][metric_name]
                    plot_executor(
                        results=result,
                        save_plot=save_plot,
                        cohort_id=_cohort_id,
                        file_name=file_name,
                    )
        else:
            result = self.result[metric_name]
            plot_executor(results=result, save_plot=save_plot, file_name=file_name)

    def plot(
        self,
        metric_name: str = "all",
        save_plot: bool = False,
        file_name: Optional[str] = None,
        cohort_id: Optional[int] = None,
    ) -> Union[str, None]:
        """
        A functon to plot the given metric

        :param metric_name: name of the metric to be plotted
        :type metric_name: str
        :param save_plot: True if the plot is to be saved
        :type save_plot: bool
        :param file_name: Plot file name
        :type file_name: str
        :param cohort_id: id of the data cohort if applicable
        :type cohort_id: int

        :return: path of the plot if saved
        :rtype: str
        """
        if metric_name == "all":
            for _metric in plot_manager.get_names(
                problem_type=self.user_params.problem_type
            ):
                _plot_executor = plot_manager[
                    f"{self.user_params.problem_type}.{_metric}"
                ]
                self._run_plot_executor(
                    _plot_executor, _metric, save_plot, cohort_id, file_name
                )
        else:
            _plot_executor = plot_manager[
                f"{self.user_params.problem_type}.{metric_name}"
            ]
            self._run_plot_executor(
                _plot_executor, metric_name, save_plot, cohort_id, file_name
            )



import os
from typing import Union, Dict, List, Optional

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from sklearn.metrics import (
    confusion_matrix,
    roc_auc_score,
    accuracy_score,
    f1_score,
    precision_score,
    recall_score,
    matthews_corrcoef,
)

from gesund.core import metric_manager, plot_manager


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(self, data: dict, get_logits=False):
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth = [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])
        return (np.asarray(prediction), np.asarray(ground_truth))

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        class_order = [int(i) for i in class_mapping.keys()]
        class_labels = [class_mapping[str(k)] for k in class_order]

        true = np.array(data["ground_truth"])
        pred_categorical = np.array(data["prediction"])
        pred_logits = data.get("pred_logits", None)

        if len(class_order) > 2:
            prediction, ground_truth = self._preprocess(data, get_logits=True)
            pred_logits = prediction
            pred_categorical = np.argmax(prediction, axis=1)
        else:
            prediction, ground_truth = self._preprocess(data)
            pred_categorical = prediction
            pred_logits = None

        true = ground_truth
        cm = confusion_matrix(true, pred_categorical, labels=class_order)
        per_class_metrics = {}
        for idx, _cls in enumerate(class_order):
            tp = cm[idx, idx]
            fn = cm[idx, :].sum() - tp
            fp = cm[:, idx].sum() - tp
            tn = cm.sum() - (tp + fp + fn)

            sensitivity = tp / (tp + fn) if (tp + fn) > 0 else 0
            specificity = tn / (tn + fp) if (tn + fp) > 0 else 0
            precision = tp / (tp + fp) if (tp + fp) > 0 else 0
            f1 = (
                2 * precision * sensitivity / (precision + sensitivity)
                if (precision + sensitivity) > 0
                else 0
            )
            mcc = matthews_corrcoef(
                (true == _cls).astype(int), (pred_categorical == _cls).astype(int)
            )

            per_class_metrics[class_mapping[str(_cls)]] = {
                "TP": tp,
                "TN": tn,
                "FP": fp,
                "FN": fn,
                "Sensitivity": sensitivity,
                "Specificity": specificity,
                "Precision": precision,
                "F1 Score": f1,
                "Matthews CC": mcc,
            }

        # Calculate overall metrics
        overall_accuracy = accuracy_score(true, pred_categorical)
        macro_f1 = f1_score(true, pred_categorical, average="macro")
        macro_precision = precision_score(true, pred_categorical, average="macro")
        macro_recall = recall_score(true, pred_categorical, average="macro")
        macro_specificity = np.mean(
            [metrics["Specificity"] for metrics in per_class_metrics.values()]
        )
        macro_mcc = matthews_corrcoef(true, pred_categorical)

        if pred_logits is not None:
            try:
                macro_auc = roc_auc_score(
                    true, pred_logits, multi_class="ovo", average="macro"
                )
            except ValueError:
                macro_auc = "Undefined"
        else:
            macro_auc = "Undefined"

        overall_metrics = {
            "Accuracy": overall_accuracy,
            "Macro F1 Score": macro_f1,
            "Macro Precision": macro_precision,
            "Macro Recall": macro_recall,
            "Macro Specificity": macro_specificity,
            "Matthews CC": macro_mcc,
            "Macro AUC": macro_auc,
        }

        result = {
            "per_class_metrics": per_class_metrics,
            "overall_metrics": overall_metrics,
            "confusion_matrix": cm,
            "classes": class_labels,
        }

        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the lift chart for the given data.

        :param data: The input data required for calculation and plotting
                     {"prediction":, "ground_truth": , "metadata":, "class_mappings":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate the result
        result = self.__calculate_metrics(data, data.get("class_mapping"))
        return result


class PlotStatsTables:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.per_class_metrics = data["per_class_metrics"]
        self.overall_metrics = data["overall_metrics"]
        self.confusion_matrix = data["confusion_matrix"]
        self.classes = data["classes"]
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting the stats tables.
        """
        required_keys = [
            "per_class_metrics",
            "overall_metrics",
            "confusion_matrix",
            "classes",
        ]
        for key in required_keys:
            if key not in self.data:
                raise ValueError(f"Data must contain '{key}'.")

    def save(self, figs: List[Figure], filenames: List[str]) -> List[str]:
        """
        Saves multiple Matplotlib Figure objects to files.

        :param figs: List of Matplotlib Figure objects to save
        :type figs: List[Figure]
        :param filenames: List of filenames where the plot images will be saved
        :type filenames: List[str]

        :return: List of file paths where the plot images are saved
        :rtype: List[str]
        """
        filepaths = []
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        for fig, filename in zip(figs, filenames):
            if self.cohort_id:
                filepath = f"{dir_path}/{self.cohort_id}_{filename}"
            else:
                filepath = f"{dir_path}/{filename}"

            fig.savefig(filepath, format="png")
            filepaths.append(filepath)

        return filepaths

    def plot(self) -> List[Figure]:
        """
        Plots the stats tables and returns the figure objects.

        :return: List of Matplotlib Figure objects [confusion_matrix, per_class_metrics]
        :rtype: List[Figure]
        """
        sns.set_style("whitegrid")

        # validate data
        self._validate_data()
        figures = []

        # Plot confusion matrix
        fig_cm, ax_cm = plt.subplots(figsize=(8, 6))
        cm = self.confusion_matrix
        sns.heatmap(
            cm,
            annot=True,
            fmt="d",
            cmap="Blues",
            xticklabels=self.classes,
            yticklabels=self.classes,
            ax=ax_cm,
        )
        ax_cm.set_xlabel(
            "Predicted Labels", fontdict={"fontsize": 14, "fontweight": "medium"}
        )
        ax_cm.set_ylabel(
            "True Labels", fontdict={"fontsize": 14, "fontweight": "medium"}
        )

        if self.cohort_id:
            title_str = f"Confusion Matrix : cohort - {self.cohort_id}"
        else:
            title_str = "Confusion Matrix"

        ax_cm.set_title(title_str)
        figures.append(fig_cm)

        # Plot per-class metrics
        metrics_df = pd.DataFrame(self.per_class_metrics).T
        metrics_to_plot = ["Sensitivity", "Specificity", "Precision", "F1 Score"]
        metrics_df = metrics_df[metrics_to_plot].reset_index()
        metrics_df = metrics_df.melt(
            id_vars="index", var_name="Metric", value_name="Value"
        )

        fig_metrics, ax_metrics = plt.subplots(figsize=(10, 6))
        sns.barplot(
            data=metrics_df,
            x="index",
            y="Value",
            hue="Metric",
            palette="Set2",
            ax=ax_metrics,
        )
        ax_metrics.set_xlabel(
            "Classes", fontdict={"fontsize": 14, "fontweight": "medium"}
        )
        ax_metrics.set_ylabel(
            "Metric Value", fontdict={"fontsize": 14, "fontweight": "medium"}
        )

        if self.cohort_id:
            title_str = f"Per-Class Metrics : cohort - {self.cohort_id}"
        else:
            title_str = "Per-Class Metrics"

        ax_metrics.set_title(
            title_str, fontdict={"fontsize": 16, "fontweight": "medium"}
        )
        ax_metrics.legend(title="Metric")
        figures.append(fig_metrics)

        # Display overall metrics
        # print("Overall Metrics:")
        # for metric, value in self.overall_metrics.items():
        #     print(
        #         f"{metric}: {value:.4f}"
        #         if isinstance(value, float)
        #         else f"{metric}: {value}"
        #     )
        return figures


class SemanticSegmentation(Classification):
    pass


class ObjectDetection(Classification):
    pass


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("classification.stats_tables")
def calculate_stats_tables(data: dict, problem_type: str):
    """
    Calculates the stats tables metric.
    """
    metric_calculator = problem_type_map[problem_type]()
    result = metric_calculator.calculate(data)
    return result


@plot_manager.register("classification.stats_tables")
def plot_stats_tables(
    results: dict, save_plot: bool, file_name: str
) -> Union[List[str], List[Figure]]:
    """
    Plots the stats tables.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plots
    :type save_plot: bool
    :param file_name: name of the files
    :type file_name: str

    :return: List of figure objects or paths to the saved plots
    :rtype: Union[List[str], List[Figure]]
    """
    plotter = PlotStatsTables(data=results)
    figs = plotter.plot()
    if save_plot:
        filenames = [f"{file_name}_{i}.png" for i in range(len(figs))]
        return plotter.save(figs, filenames=filenames)
    else:
        plt.show()



from typing import Union, Optional
import os

import numpy as np
import pandas as pd
from sklearn.metrics import auc, roc_curve
from sklearn.preprocessing import label_binarize
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        Validates the data required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": }
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        # check for image ids samples in the ground truth and prediction
        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground truth samples does not match.")

        return True

    def _preprocess(self, data: dict, get_logits=False) -> tuple:
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth = [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])
        return (np.asarray(prediction), np.asarray(ground_truth))

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        class_order = [int(i) for i in list(class_mapping.keys())]

        # Calculate ROC and AUC
        # TODO: class wise auc and roc

        if len(class_order) > 2:
            # implementation of multi class classification
            # logits are required for multi class classification
            prediction, ground_truth = self._preprocess(data, get_logits=True)
            # TODO: multi class auc - roc calculation

        else:
            prediction, ground_truth = self._preprocess(data)
            fpr, tpr, thresholds = roc_curve(prediction, ground_truth)
            auc_score = auc(fpr, tpr)
            fpr = [float(round(value, 4)) for value in fpr]
            tpr = [float(round(value, 4)) for value in tpr]

        result = {
            "fpr": fpr,
            "tpr": tpr,
            "auc": auc_score,
            "class_mapping": class_mapping,
            "class_order": class_order,
            "thresholds": thresholds,
        }

        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the AUC metric for the given dataset.

        :param data: The input data required for calculation and plotting
                     {"prediction":, "ground_truth": , "class_mappings":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate results
        result = self.__calculate_metrics(data, data.get("class_mapping"))

        return result


class SemanticSegmentation(Classification):
    pass


class ObjectDetection:
    pass


class PlotAuc:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.fpr = data["fpr"]
        self.tpr = data["tpr"]
        self.aucs = data["auc"]
        self.class_mappings = data["class_mapping"]
        self.class_order = data["class_order"]
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting the AUC.
        """
        required_keys = ["fpr", "tpr", "auc"]
        for key in required_keys:
            if key not in self.data:
                raise ValueError(f"Data must contain '{key}'.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file.

        :param filename: Path where the plot image will be saved
        :type filename: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"
        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        Plots the AUC curves.
        """
        sns.set_style("whitegrid")
        # Validate the data
        self._validate_data()

        fig, ax = plt.subplots(figsize=(10, 7))
        # TODO: Class wise plot auc-roc
        sns.lineplot(x=self.fpr, y=self.tpr, ax=ax)
        ax.set_xlabel(
            "False Positive Rate", fontdict={"fontsize": 14, "fontweight": "medium"}
        )
        ax.set_ylabel(
            "True Positive Rate", fontdict={"fontsize": 14, "fontweight": "medium"}
        )
        if self.cohort_id:
            title_str = f"Receiver Operating Characteristic (ROC) Curve : cohort - {self.cohort_id}"
        else:
            title_str = "Receiver Operating Characteristic (ROC) Curve"

        ax.set_title(
            title_str,
            fontdict={"fontsize": 16, "fontweight": "medium"},
        )
        ax.legend(loc="lower right")
        return fig


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("classification.auc")
def calculate_auc_metric(data: dict, problem_type: str):
    """
    A wrapper function to calculate the AUC metric.

    :param data: Dictionary of data: {"prediction": , "ground_truth": , 'metadata': , 'class_mapping': }
    :type data: dict
    :param problem_type: Type of the problem
    :type problem_type: str

    :return: Dict of calculated results
    :rtype: dict
    """
    _metric_calculator = problem_type_map[problem_type]()
    result = _metric_calculator.calculate(data)
    return result


@plot_manager.register("classification.auc")
def plot_auc(
    results: dict,
    save_plot: bool,
    file_name: str = "auc.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the AUC curves.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plot
    :type save_plot: bool
    :param file_name: name of the file
    :type file_name: str
    :param cohort_id: id of the cohort
    :type cohort_id: int

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotAuc(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



class CustomError(Exception):
    def __init__(self, msg: str):
        self.message = msg
        super().__init__(msg)

    def __str__(self):
        return self.message


class InputError(CustomError):
    pass


class DataLoadError(CustomError):
    pass


class MetricCalculationError(CustomError):
    pass


class PlotError(CustomError):
    pass


class FunctionNotFoundError(CustomError):
    pass


class RegistrationNotAllowed(CustomError):
    pass



import json
import pandas as pd

from ._exceptions import DataLoadError


class DataLoader:
    def __init__(self, data_format: str):
        """
        A function to initialize the data

        :param data_format: string value indicating the data
        :type data_format: str

        :return: None
        """
        data_loaders = {"csv": self._csv_loader, "json": self._json_loader}
        self._loader_fxn = data_loaders[data_format]

    @staticmethod
    def _json_loader(src_path: str) -> list:
        """
        A function to load JSON file

        :param src_path: source path of the file
        :type src_path: str

        :return: loaded data
        :rtype: list
        """
        try:
            with open(src_path, "r") as f:
                data = json.load(f)
                return data
        except Exception as e:
            print(e)
            raise DataLoadError("Could not load JSON file !")

    @staticmethod
    def _csv_loader(src_path: str) -> pd.DataFrame:
        """
        A function to load csv file

        :param src_path: source path of the file
        :type src_path: str

        :return: loaded data frame
        :rtype: pd.DataFrame
        """
        return pd.DataFrame(src_path)

    def load(self, src_path: str) -> dict:
        """
        A function to load the data

        :param src_path: source path of the file
        :type src_path: str

        :return: None
        """
        return self._loader_fxn(src_path)



import os
from typing import Union, Dict, List, Optional

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.pyplot import Figure
from sklearn.metrics import (
    f1_score,
    precision_score,
    recall_score,
    confusion_matrix,
    matthews_corrcoef,
    roc_curve,
)
import seaborn as sns

from gesund.core import metric_manager, plot_manager


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(self, data: dict, get_logits=False):
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth = [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])
        return (np.asarray(prediction), np.asarray(ground_truth))

    @staticmethod
    def _apply_threshold(
        threshold: float, y_true: np.ndarray, y_pred: np.ndarray
    ) -> pd.DataFrame:
        """
        A function to apply threshold

        :param threshold: The threshold value to be applied
        :type threshold: float
        :param y_true: the ground truth values in numpy array
        :type y_true: np.ndarray
        :param y_pred: the probabilites in numpy array
        :type y_pred: np.ndarray

        :return: results calculated and composed in dataframe
        :rtype: pd.DataFrame
        """
        predicted = (y_pred >= threshold).astype(int)

        # calculate the confusion matrix
        tn, fp, fn, tp = confusion_matrix(y_true, predicted).ravel()

        # calculate the metrics
        metrics = {}
        metrics["f1_score"] = f1_score(y_true, predicted)
        metrics["precision"] = precision_score(y_true, predicted)
        metrics["sensitivity"] = recall_score(y_true, predicted)
        metrics["specificity"] = round(tn / (tn + fp), 4)
        metrics["mcc"] = matthews_corrcoef(y_true, predicted)
        metrics["fpr"] = round(fp / (fp + tn), 4)
        metrics["fnr"] = round(fn / (fn + tp), 4)

        metrics = {i: [metrics[i]] for i in metrics}
        metrics["threshold"] = [threshold]
        metrics = pd.DataFrame(metrics)
        return metrics

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        from gesund.core._metrics.common.top_losses import Classification

        prediction, ground_truth = self._preprocess(data, get_logits=True)
        pred_gt_df = pd.DataFrame(prediction)

        # softmax is applied to make the logits aka the raw outputs of the model
        # comparable and to sum upto 1
        probabilities = Classification._softmax(prediction)
        pred_gt_df["probabilities"] = probabilities[:, 1]
        pred_gt_df["ground_truth"] = ground_truth

        thresholds = data["metric_args"]["threshold"]
        th_result = pd.DataFrame()
        if isinstance(thresholds, list):
            for _threshold in thresholds:
                _result = self._apply_threshold(
                    threshold=_threshold,
                    y_true=pred_gt_df["ground_truth"],
                    y_pred=pred_gt_df["probabilities"],
                )

                th_result = pd.concat([th_result, _result], axis=0, ignore_index=True)
        else:
            _result = self._apply_threshold(
                threshold=thresholds,
                y_true=pred_gt_df["ground_truth"],
                y_pred=pred_gt_df["probabilities"],
            )
            th_result = pd.concat([th_result, _result], axis=0, ignore_index=True)

        result = {"threshold_result": th_result, "pred_gt_data": pred_gt_df}
        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the top losses for the given dataset.

        :param data: The input data required for calculation and plotting
                {"prediction":, "ground_truth": , "metadata":, "class_mappings":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self.__calculate_metrics(data, data.get("class_mapping"))

        return result


class PlotThreshold:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting the threshold.
        """
        required_keys = ["threshold_result"]
        for key in required_keys:
            if key not in self.data:
                raise ValueError(f"Data must contain '{key}'.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file.

        :param filename: Path where the plot image will be saved
        :type filename: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        Plots the ROC Curve with the optimal threshold highlighted.
        """
        sns.set_style("whitegrid")

        self._validate_data()
        fig, ax = plt.subplots(figsize=(10, 7))
        plot_data = pd.melt(
            self.data["threshold_result"],
            id_vars=["threshold"],
            var_name="metric",
            value_name="value",
        )

        sns.barplot(
            x="metric",
            y="value",
            hue="threshold",
            data=plot_data,
            ax=ax,
            palette="pastel",
        )
        ax.set_xlabel("Metric", fontdict={"fontsize": 14, "fontweight": "medium"})
        ax.set_ylabel("Value", fontdict={"fontsize": 14, "fontweight": "medium"})

        if self.cohort_id:
            title_str = f"Threshold adjusted metrics : cohort - {self.cohort_id}"
        else:
            title_str = "Threshold adjusted metrics"

        ax.set_title(
            title_str,
            fontdict={"fontsize": 16, "fontweight": "medium"},
        )
        ax.legend(loc="lower right")
        return fig


class SemanticSegmentation(Classification):
    pass


class ObjectDetection:
    pass


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("classification.threshold")
def calculate_threshold_metric(data: dict, problem_type: str):
    """
    A wrapper function to calculate the threshold metric.

    :param data: Dictionary of data: {"prediction": , "ground_truth": }
    :type data: dict
    :param problem_type: Type of the problem
    :type problem_type: str

    :return: Calculated results
    :rtype: dict
    """
    metric_calculator = problem_type_map[problem_type]()
    result = metric_calculator.calculate(data)
    return result


@plot_manager.register("classification.threshold")
def plot_threshold(
    results: dict, save_plot: bool, file_name: str = "threshold.png"
) -> Union[str, None]:
    """
    A wrapper function to plot the threshold.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plot
    :type save_plot: bool
    :param file_name: Name of the file
    :type file_name: str

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotThreshold(data=results)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



from typing import Optional, Union
import os

from tqdm import tqdm
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager
from .iou import IoUCalc


class Classification:
    pass


class ObjectDetection:
    pass


class SemanticSegmentation:
    def __init__(self):
        self.iou = IoUCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a boolean value
        :rtype: bool
        """
        check_keys = ("ground_truth", "prediction")
        for _key in check_keys:
            if _key not in data:
                raise ValueError(f"Missing {_key} in the data dictionary")

        # check the common set of images
        common_ids = set(list(data["prediction"].keys())).difference(
            set(list(data["ground_truth"].keys()))
        )

        if common_ids:
            raise ValueError(
                "prediction and ground truth does not have corresponding samples"
            )

    @staticmethod
    def convert_rle_to_boxes(rle_str: str, width: int, height: int) -> list:
        """
        A function to convert the RLE string to boxes

        :param rle_str: the RLE string
        :type rle_str: str
        :param width: the width of the image
        :type width: int
        :param height: the height of the image
        :type height: int

        :return: a list containing the boxes
        :rtype: list
        """
        rle_nums = list(map(int, rle_str.split()))
        rle_pairs = list(zip(rle_nums[::2], rle_nums[1::2]))

        mask = np.zeros(height * width, dtype=np.uint8)
        for start, length in rle_pairs:
            mask[start : start + length] = 1

        mask = mask.reshape((height, width)).T
        bboxes = []
        top_left, bottom_right = None, None

        # find bounding boxes
        for i in range(height):
            for j in range(width):
                if mask[i, j] == 1:
                    top_left = (i, j)
                    break

        for i in range(height - 1, -1, -1):
            for j in range(width - 1, -1, -1):
                if mask[i, j] == 1:
                    bottom_right = (i, j)
                    break

        if top_left and bottom_right:
            bboxes = [top_left[1], top_left[0], bottom_right[1], bottom_right[0]]
        return bboxes

    @staticmethod
    def convert_rle_to_mask(rle_str: str, width: int, height: int) -> np.ndarray:
        """
        A function to convert the RLE string to mask

        :param rle_str: the RLE string
        :type rle_str: str
        :param width: the width of the image
        :type width: int
        :param height: the height of the image
        :type height: int

        :return: a numpy array containing the mask
        :rtype: np.ndarray
        """
        rle_nums = list(map(int, rle_str.split()))
        rle_pairs = list(zip(rle_nums[::2], rle_nums[1::2]))

        mask = np.zeros(height * width, dtype=np.uint8)
        for start, length in rle_pairs:
            mask[start : start + length] = 1

        mask = mask.reshape((height, width)).T
        return mask

    def _preprocess(self, data: dict) -> tuple:
        """
        A function to preprocess the data

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a tuple containing the ground truth and prediction boxes
        :rtype: tuple
        """
        pred_data, gt_data = {}, {}

        for image_id in tqdm(data["ground_truth"], desc="Preprocessing data"):
            gt_masks = []
            for annotation in data["ground_truth"][image_id]["annotation"]:
                if annotation["type"] == "mask":
                    rle = annotation["mask"]["mask"]
                    width, height = annotation["shape"]
                    mask = self.convert_rle_to_mask(rle, width, height)
                    gt_masks.append(mask)
            gt_data[image_id] = gt_masks

            pred_masks = []
            width, height = data["prediction"][image_id]["shape"]
            for rle_mask in data["prediction"][image_id]["masks"]["rles"]:
                rle = rle_mask["rle"]
                mask = self.convert_rle_to_mask(rle, width, height)
                pred_masks.append(mask)
            pred_data[image_id] = pred_masks

        return (gt_data, pred_data)

    def _calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics for object detection

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        :param class_mapping: a dictionary containing the class mapping
        :type class_mapping: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        # preprocess the data
        gt_data, pred_data = self._preprocess(data)

        # calculate the metrics
        data = {"image_id": [], "iou": []}

        for image_id in gt_data:
            gt_masks = gt_data[image_id]
            if not gt_masks:
                continue
            for gt_mask in gt_masks:
                for pred_mask in pred_data[image_id]:
                    iou = self.iou.calculate_iou_mask(gt_mask, pred_mask)
                    data["image_id"].append(image_id)
                    data["iou"].append(iou)

        data = pd.DataFrame(data)
        data = data.dropna()
        return data

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the metrics for object detection

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        result = {}

        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result["iou_distribution"] = self._calculate_metrics(
            data, class_mapping=data.get("class_mapping", {})
        )

        return result


class PlotPredictedDistribution:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validation_data(self):
        """
        Validate the data
        """
        if "iou_distribution" not in self.data:
            raise ValueError("Missing iou_distribution in the data dictionary")

    def save(self, fig: Figure, filename: str) -> str:
        """
        A function to save the plot

        :param fig: the figure object
        :type fig: Figure
        :param filename: the name of the file
        :type filename: str

        :return: the filename
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        A function to plot the predicted distribution

        :return: the figure object
        :rtype: Figure
        """
        # validate the data
        self._validation_data()

        sns.set_theme(style="whitegrid")
        fig, ax = plt.subplots(figsize=(10, 5))
        plot_data = self.data["iou_distribution"]

        # plot violin plot
        sns.violinplot(x=plot_data["iou"], ax=ax)
        ax.set_xlabel("IoU", fontdict={"fontsize": 14, "fontweight": "medium"})

        if self.cohort_id:
            title_str = f"IoU distribution: cohort - {self.cohort_id}"
        else:
            title_str = "IoU distribution"

        ax.set_title(title_str, fontdict={"fontsize": 16, "fontweight": "bold"})
        return fig


problem_type_map = {
    "classification": Classification,
    "object_detection": ObjectDetection,
    "semantic_segmentation": SemanticSegmentation,
}


@metric_manager.register("semantic_segmentation.iou_distribution")
def calculate_iou_distribution(data: dict, problem_type: str) -> dict:
    """
    A function to calculate the IoU distribution

    :param data: a dictionary containing the ground truth and prediction data
    :type data: dict
    :param problem_type: the type of the problem
    :type problem_type: str

    :return: a dictionary containing the calculated metrics
    :rtype: dict
    """
    metric_calculator = problem_type_map[problem_type]()
    result = metric_calculator.calculate(data)
    return result


@plot_manager.register("semantic_segmentation.iou_distribution")
def plot_iou_dist_sem_seg(
    results: dict,
    save_plot: bool,
    file_name: str = "iou_distribution.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the threshold

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save the plot
    :type save_plot: bool
    :param file_name: the name of the file
    :type file_name: str
    :param cohort_id: the cohort id
    :type cohort_id: int

    :return: the file path
    :rtype: str
    """
    plotter = PlotPredictedDistribution(data=results)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



import os
from typing import Union, Dict, List, Optional

import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(self, data: dict, get_logits=False):
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth = [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])
        return (np.asarray(prediction), np.asarray(ground_truth))

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        class_order = [int(i) for i in class_mapping.keys()]

        if len(class_order) > 2:
            prediction, ground_truth = self._preprocess(data, get_logits=True)
        else:
            prediction, ground_truth = self._preprocess(data)

        cm = confusion_matrix(ground_truth, prediction, labels=class_order)
        confused_pairs = []
        for i, actual_class in enumerate(class_order):
            for j, pred_class in enumerate(class_order):
                if i != j and cm[i, j] > 0:
                    confused_pairs.append(
                        {
                            "true": class_mapping[str(actual_class)],
                            "predicted": class_mapping[str(pred_class)],
                            "count": cm[i, j],
                        }
                    )
        confused_pairs = sorted(confused_pairs, key=lambda x: x["count"], reverse=True)

        result = {
            "confused_pairs": confused_pairs,
            "class_mapping": class_mapping,
            "class_order": class_order,
        }
        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the lift chart for the given data.

        :param data: The input data required for calculation and plotting
                     {"prediction":, "ground_truth": , "metadata":, "class_mapping":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self.__calculate_metrics(data, data.get("class_mapping"))
        return result


class SemanticSegmentation(Classification):
    pass


class ObjectDetection:
    pass


class PlotMostConfused:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.confused_pairs = data["confused_pairs"]
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting.
        """
        if "confused_pairs" not in self.data:
            raise ValueError("Data must contain 'confused_pairs'.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file.

        :param filename: Path where the plot image will be saved
        :type filename: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self, top_k: int = 5) -> Figure:
        """
        Plots the most confused classes.

        :param top_k: Number of top confused pairs to display
        :type top_k: int

        :return: Matplotlib Figure object
        :rtype: Figure
        """
        sns.set_style("whitegrid")

        # validate the data
        self._validate_data()

        df = pd.DataFrame(self.confused_pairs[:top_k])

        fig, ax = plt.subplots(figsize=(10, 6))
        sns.barplot(x="count", y=df.index, data=df, orient="h")
        ax.set_yticks(df.index)
        ax.set_yticklabels(
            [
                f"True: {row['true']} | Predicted: {row['predicted']}"
                for _, row in df.iterrows()
            ]
        )
        ax.set_xlabel("Count", fontdict={"fontsize": 14, "fontweight": "medium"})
        ax.set_ylabel(
            "Confusion Pairs", fontdict={"fontsize": 14, "fontweight": "medium"}
        )

        if self.cohort_id:
            title_str = f"Most Confused Classes : cohort - {self.cohort_id}"
        else:
            title_str = "Most Confused Classes"

        ax.set_title(title_str, fontdict={"fontsize": 16, "fontweight": "medium"})
        fig.tight_layout()
        return fig


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("classification.most_confused")
def calculate_most_confused_metric(data: dict, problem_type: str):
    """
    Calculates the most confused classes metric.

    :param data: Dictionary of data: {"prediction": , "ground_truth": }
    :type data: dict
    :param problem_type: Type of the problem
    :type problem_type: str

    :return: Calculated results
    :rtype: dict
    """
    metric_calculator = problem_type_map[problem_type]()
    result = metric_calculator.calculate(data)
    return result


@plot_manager.register("classification.most_confused")
def plot_most_confused(
    results: dict,
    save_plot: bool,
    file_name: str = "most_confused.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the lift chart.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plot
    :type save_plot: bool
    :param file_name: Name of the file to save the plot
    :type file_name: str
    :param cohort_id: id of the cohort
    :type cohort_id: int

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotMostConfused(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



from typing import Union, Optional
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager
from gesund.core._utils import ValidationUtils
from .iou import IoUCalc


class Classification:
    pass


class SemanticSegmentation:
    def __init__(self):
        pass

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a boolean value
        :rtype: bool
        """

        return ObjectDetection._validate_data(ObjectDetection, data)

    def _preprocess(self, data: dict, get_class_only=False) -> tuple:
        """
        A function to preprocess the data

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        ;param get_class_only: a flag to get only the class labels
        :type get_class_only: bool

        :return: a tuple containing the ground truth and prediction boxes
        :rtype: tuple
        """
        results = {"gt_class_label": [], "pred_class_label": []}
        for image_id in data["ground_truth"]:
            for annotation in data["ground_truth"][image_id]["annotation"]:
                if annotation["type"] == "mask":
                    results["gt_class_label"].append(annotation["label"])

            for pred_mask in data["prediction"][image_id]["masks"]["rles"]:
                results["pred_class_label"].append(pred_mask["class"])

        results = pd.DataFrame(results)
        return results

    def _calculate_metrics(self, data: dict) -> dict:
        """
        A function to calculate the predicted_distribution metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        results = {}
        # preprocess the data
        label_dist = self._preprocess(data)
        results["label_distribution"] = label_dist
        return results

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the predicted_distribution metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        result = {}
        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self._calculate_metrics(data)

        return result


class ObjectDetection:
    def __init__(self):
        self.iou = IoUCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        """
        # check for the important keys in the data
        check_keys = ("ground_truth", "prediction")
        for _key in check_keys:
            if _key not in data:
                raise ValueError(f"Missing {_key} in the data dictionary")

        # check the common set of images
        common_ids = set(list(data["prediction"].keys())).difference(
            set(list(data["ground_truth"].keys()))
        )

        if common_ids:
            raise ValueError(
                "prediction and ground truth does not have corresponding samples"
            )

    def _preprocess(self, data: dict) -> tuple:
        """
        A function to preprocess the data

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a tuple containing the ground truth and prediction boxes
        :rtype: tuple
        """
        from .average_precision import ObjectDetection

        return ObjectDetection._preprocess(data, get_label=True)

    def _organize_data(
        self, gt_boxes: dict, pred_boxes: dict, class_mapping: dict
    ) -> dict:
        """
        A function to organize the data

        :param gt_boxes: a dictionary containing the ground truth boxes
        :type gt_boxes: dict
        :param pred_boxes: a dictionary containing the prediction boxes
        :type pred_boxes: dict
        :param class_mapping: a dictionary containing the class mapping
        :type class_mapping: dict

        :return: a dictionary containing the organized data
        :rtype: dict
        """
        data = {"pred_class_label": [], "gt_class_label": []}

        for image_id in gt_boxes:
            data["gt_class_label"].extend([box[-1] for box in gt_boxes[image_id]])
            data["pred_class_label"].extend([box[-1] for box in pred_boxes[image_id]])

        return data

    def _calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the predicted_distribution metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        :param class_mapping: a dictionary containing the class mapping
        :type class_mapping: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        results = {}
        # preprocess the data
        gt_boxes, pred_boxes = self._preprocess(data)

        label_dist = self._organize_data(gt_boxes, pred_boxes, class_mapping)

        results["label_distribution"] = label_dist

        return results

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the predicted_distribution metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        result = {}
        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self._calculate_metrics(data, data.get("class_mapping"))

        return result


class PlotPredictedDistribution:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validate the data
        """
        if "label_distribution" not in self.data:
            raise ValueError("Missing label_distribution in the data dictionary")

    def save(self, fig: Figure, filename: str) -> str:
        """
        A function to save the plot

        :param fig: the figure object
        :type fig: Figure
        :param filename: the name of the file
        :type filename: str

        :return: the path to the saved file
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        A function to plot the predicted distribution

        :return: the figure object
        :rtype: Figure
        """
        # validate the data
        self._validate_data()

        sns.set_theme(style="whitegrid")
        figx, ax = plt.subplots(1, 2, figsize=(10, 6), subplot_kw=dict(aspect="equal"))
        colors1 = sns.color_palette("pastel")
        colors2 = sns.color_palette("pastel")
        plot_data = self.data["label_distribution"]
        plot_data = pd.DataFrame(plot_data)

        # plot first donut chart
        value_counts1 = plot_data["gt_class_label"].value_counts()
        wedges1, texts1, autotexts1 = ax[0].pie(
            value_counts1,
            labels=value_counts1.index,
            autopct="%1.1f%%",
            colors=colors1,
            startangle=90,
            wedgeprops=dict(width=0.3),
        )
        ax[0].add_artist(plt.Circle((0, 0), 0.7, fc="white"))

        if self.cohort_id:
            title_str = f"Ground Truth Distribution (Cohort ID: {self.cohort_id})"
        else:
            title_str = "Ground Truth Distribution"

        ax[0].set_title(title_str)
        ax[0].legend(
            wedges1,
            value_counts1.index,
            title="Classes",
            loc="center left",
            bbox_to_anchor=(1, 0, 0.5, 1),
        )

        # plot second donut chart
        value_counts2 = plot_data["pred_class_label"].value_counts()
        wedges2, texts2, autotexts2 = ax[1].pie(
            value_counts2,
            labels=value_counts2.index,
            autopct="%1.1f%%",
            colors=colors2,
            startangle=90,
            wedgeprops=dict(width=0.3),
        )

        ax[1].add_artist(plt.Circle((0, 0), 0.7, fc="white"))

        if self.cohort_id:
            title_str = f"Predicted Distribution (Cohort ID: {self.cohort_id})"
        else:
            title_str = "Predicted Distribution"

        ax[1].set_title(title_str)
        ax[1].legend(
            wedges2,
            value_counts2.index,
            title="Classes",
            loc="center left",
            bbox_to_anchor=(1, 0, 0.5, 1),
        )

        plt.suptitle("Label Distribution")

        return figx


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("semantic_segmentation.predicted_distribution")
@metric_manager.register("object_detection.predicted_distribution")
def calculate_predicted_distribution(data: dict, problem_type: str):
    """
    A wrapper function to calculate the predicted_distribution metrics.

    :param data: a dictionary containing the ground truth and prediction data
    :type data: dict
    :param problem_type: the type of problem
    :type problem_type: str

    :return: calculated metric results
    :rtype: dict
    """
    _metric_calculator = problem_type_map[problem_type]()
    result = _metric_calculator.calculate(data)
    return result


@plot_manager.register("semantic_segmentation.predicted_distribution")
@plot_manager.register("object_detection.predicted_distribution")
def plot_predicted_distribution_od(
    results: dict,
    save_plot: bool,
    file_name: str = "predicted_distribution.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the predicted distribution metrics.

    :param results: a dictionary containing the results of the predicted distribution
    :type results: dict
    :param save_plot: a flag to save the plot
    :type save_plot: bool
    :param file_name: the name of the file
    :type file_name: str
    :param cohort_id: the cohort id
    :type cohort_id: Optional[int]

    :return: the path to the saved file
    :rtype: Union[str, None]
    """
    plotter = PlotPredictedDistribution(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



from typing import Optional

import numpy as np


class DiceCalc:
    def __init__(self):
        pass

    @staticmethod
    def dice_loss(y_true: np.ndarray, y_pred: np.ndarray, smooth: float = 1e-6):
        """
        Function to calculate the dice loss

        :param y_true: The true labels as a 4D array (batch_size, height, width, num_classes)
        :type y_true: np.ndarray
        :param y_pred: The predicted labels as a 4d array (batch_size, height, width, num_classes)
        """
        y_true_flat = y_true.flatten()
        y_pred_flat = y_pred.flatten()
        intersection = np.sum(y_true_flat * y_pred_flat)
        union = np.sum(y_true_flat + y_pred_flat)
        dice_score = (2.0 * intersection + smooth) / (union + smooth)
        return round(1.0 - np.mean(dice_score), 4)

    def calculate(self, mask1: np.ndarray, mask2: np.ndarray) -> float:
        """
        Calculate the dice coefficient between two masks

        :param mask1: the first mask
        :type mask1: np.ndarray
        :param mask2: the second mask
        :type mask2: np.ndarray

        :return: the dice coefficient
        :rtype: float
        """
        intersection = np.sum(mask1 * mask2)
        sum_masks = np.sum(mask1) + np.sum(mask2)
        dice = 2.0 * intersection / sum_masks
        return dice


import os
from pydantic import BaseModel, field_validator
from typing import List, Dict, Union, Optional, ClassVar

from ._exceptions import InputError


class UserInputParams(BaseModel):
    annotations_path: str
    predictions_path: str
    class_mapping: Union[str, dict]
    problem_type: str
    json_structure_type: str
    data_format: str
    plot_config: dict
    metadata_path: Optional[str] = None

    allowed_values: ClassVar[dict] = {
        "problem_type": ["classification", "object_detection", "semantic_segmentation"],
        "json_structure_type": ["gesund", "coco", "yolo"],
        "data_format": ["json"],
    }

    @field_validator("annotations_path")
    def validate_annotations_path(cls, annotations_path):
        if os.path.exists(annotations_path):
            print("annotations path validated !")
        else:
            raise InputError(msg="Annotations path is invalid")
        return annotations_path

    @field_validator("predictions_path")
    def validate_predictions_path(cls, predictions_path):
        if os.path.exists(predictions_path):
            print("predictions path validated !")
        else:
            raise InputError(msg="Predictions path is invalid")
        return predictions_path

    @field_validator("metadata_path")
    def validate_metadata_path(cls, metadata_path):
        if metadata_path:
            if os.path.exists(metadata_path):
                print("metadata path validated !")
            else:
                raise InputError(msg="Metadata path is invalid")
        else:
            print("No metadata path provided.")
        return metadata_path

    @field_validator("class_mapping")
    def validate_class_mapping(cls, class_mapping):
        if isinstance(class_mapping, str):
            if os.path.exists(class_mapping):
                print("class mapping file validated !")
            else:
                raise InputError(msg="Class mapping path is invalid")
        else:
            print("Class mapping is dict input")
        return class_mapping

    @field_validator("problem_type")
    def validate_problem_type(cls, problem_type):
        if problem_type not in cls.allowed_values["problem_type"]:
            raise InputError("Invalid problem type")
        else:
            print("Problem Type validated !")
        return problem_type

    @field_validator("json_structure_type")
    def validate_json_structure_type(cls, json_structure_type):
        if json_structure_type not in cls.allowed_values["json_structure_type"]:
            raise InputError("Invalid json structure type")
        else:
            print("JSON structure type validated!")
        return json_structure_type

    @field_validator("data_format")
    def validate_data_format(cls, data_format):
        if data_format not in cls.allowed_values["data_format"]:
            raise InputError("Invalid data format")
        else:
            print("Data format validated!")
        return data_format


class UserInputData(BaseModel):
    prediction: Union[List[Dict], Dict]
    annotation: Union[List[Dict], Dict]
    class_mapping: Dict
    metadata: Optional[Union[List[Dict], Dict]] = None
    converted_prediction: Optional[Union[List[Dict], Dict]] = None
    converted_annotation: Optional[Union[List[Dict], Dict]] = None
    was_converted: Optional[bool] = False



from typing import Callable, Dict, TypeVar, Generic, List, Optional

from gesund.core._exceptions import FunctionNotFoundError, RegistrationNotAllowed

T = TypeVar("T")


class GenericPMManager(Generic[T]):
    """
    A generic class manager for plot and metric manager

    """

    def __init__(self):
        self._store: Dict[str, T] = {}
        self._allowed_registerations = [
            "classification",
            "semantic_segmentation",
            "instance_segmentation",
            "object_detection",
        ]

    def register(self, name: str) -> Callable[[T], T]:
        """
        A decorator function to register a function under a given name:

        :param name: the name to register the function under
        :type name: str

        :return: The decorator function
        :rtype: Callable
        """

        def decorator(func: T) -> T:
            if "." in name:
                _k1, _k2 = name.split(".")
                if _k1 not in self._allowed_registerations:
                    raise RegistrationNotAllowed(
                        f"{_k1} not allowed, only allowed {','.join(self._allowed_registerations)}"
                    )

                if _k1 in self._store:
                    self._store[_k1][_k2] = func
                else:
                    self._store[_k1] = {}
                    self._store[_k1][_k2] = func
            else:
                self._store[name] = func
            return func

        return decorator

    def get_names(self, problem_type: Optional[str] = None) -> List[str]:
        """
        Get the names of the registered functions,

        :param problem_type: name of the problem type
        :type problem_type:str

        :return: A list of registered names.
        """
        if problem_type:
            return list(self._store[problem_type].keys())
        return list(self._store.keys())

    def __getitem__(self, key: str) -> Callable:
        """
        Get the metric function given the metric name

        :param key: The metric / plot name
        :type key: str

        :return: the metric function associated with the given name
        :raises FunctionNotFoundError: if the registration key is not found
        :rtype: Callable
        """
        try:
            if "." in key:
                _k1, _k2 = key.split(".")
                return self._store[_k1][_k2]
            else:
                return self._store[key]
        except Exception as e:
            print(e)
            raise FunctionNotFoundError(f"{key} not registered")



from typing import Optional


import numpy as np


class IoUCalc:
    def __init__(self):
        pass

    def calculate_iou_loss(self, gt_boxes: list, pred_boxes: list):
        """
        Calculates the intersection over union loss between ground truth and predicted box

        :param gt_boxes: list of ground truth boxes
        :type gt_boxes: list
        :param pred_boxes: list of predicted boxes
        :type pred_boxes: list

        :return: iou loss
        :rtype: float
        """
        num_gt_boxes = len(gt_boxes)
        num_pred_boxes = len(pred_boxes)

        iou_loss = 0.0

        for i in range(num_gt_boxes):
            max_iou = 0.0
            for j in range(num_pred_boxes):
                iou = self._calculate_box(gt_boxes[i], pred_boxes[j])
                max_iou = max(max_iou, iou)
            iou_loss += (1 - max_iou) ** 2

        iou_loss /= num_gt_boxes
        return iou_loss

    def _calculate_box(self, box1, box2):
        """
        A function to calculate the bounding box

        :param box1:
        :type box1:
        :param box2:
        :type box2:

        :return: the calculated Iou value
        :rtype: float
        """
        x1_i = np.maximum(box1[0], box2[0])
        y1_i = np.maximum(box1[1], box2[1])
        x2_i = np.maximum(box1[2], box2[2])
        y2_i = np.maximum(box1[3], box2[3])

        inter_area = np.maximum(0, x2_i - x1_i) * np.maximum(0, y2_i - y1_i)

        box1_area = (box1[2] - box1[0]) * (box1[3] - box1[1])
        box2_area = (box2[2] - box2[0]) * (box2[3] - box2[1])

        union_area = box1_area + box2_area - inter_area

        iou = inter_area / union_area
        return iou

    def calculate(self, y_true: np.ndarray, y_pred: np.ndarray):
        return self._calculate_box(y_true, y_pred)

    @staticmethod
    def calculate_iou_mask(mask1, mask2):
        """
        A function to calculate the intersection over union between two masks

        :param mask1: the first mask
        :type mask1: np.ndarray
        :param mask2: the second mask
        :type mask2: np.ndarray

        :return: the calculated iou value
        :rtype: float
        """
        if mask1.shape != mask2.shape:
            raise ValueError("Masks must have the same shape.")

        intersection = np.logical_and(mask1, mask2)
        union = np.logical_or(mask1, mask2)

        iou = np.sum(intersection) / np.sum(union)
        return round(iou, 4)



from typing import Optional, Union
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager
from .dice import DiceCalc
from .iou import IoUCalc


class Classification:
    pass


class ObjectDetection:
    pass


class SemanticSegmentation:
    def __init__(self):
        self.iou = IoUCalc()
        self.dice = DiceCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a boolean value
        :rtype: bool
        """
        from .iou_distribution import SemanticSegmentation

        return SemanticSegmentation._validate_data(SemanticSegmentation, data)

    def _preprocess(self, data: dict) -> dict:
        """
        A function to preprocess the data

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the preprocessed data
        :rtype: dict
        """
        from .iou_distribution import SemanticSegmentation

        return SemanticSegmentation._preprocess(SemanticSegmentation, data)

    def _calculate_metrics(self, data: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        # preprocess the data
        gt_data, pred_data = self._preprocess(data)

        # calculate the dice distribution
        result = {"image_id": [], "dice": [], "iou": []}

        for image_id in gt_data:
            gt_masks = gt_data[image_id]
            if not gt_masks:
                continue

            for gt_mask in gt_masks:
                for pred_mask in pred_data[image_id]:
                    dice = self.dice.calculate(gt_mask, pred_mask)
                    iou = self.iou.calculate_iou_mask(gt_mask, pred_mask)
                    result["image_id"].append(image_id)
                    result["dice"].append(dice)
                    result["iou"].append(iou)

            result = pd.DataFrame(result)
            result = result.dropna()
            return result

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the dice distribution for semantic segmentation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the dice distribution
        :rtype: dict
        """
        result = {}

        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result["dice_distribution"] = self._calculate_metrics(data)

        return result


class PlotDiceDistribution:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        A function to validate the data
        """
        if "dice_distribution" not in self.data:
            raise ValueError("Missing dice_distribution in the data dictionary")

    def save(self, fig: Figure, filename: str) -> str:
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        A function to plot the predicted distribution

        :return: the figure object
        :rtype: Figure
        """
        # validate the data
        self._validate_data()

        sns.set_theme(style="whitegrid")
        plot_data = self.data["dice_distribution"]

        g = sns.JointGrid(data=plot_data, x="dice", y="iou", space=0, height=9, ratio=7)

        g.plot_joint(sns.scatterplot, palette="pastel")
        g.plot_marginals(sns.histplot, kde=True, color=".5")

        g.set_axis_labels("DICE", "IoU", fontsize=14)
        g.figure.suptitle("Scatterplot DICE vs IoU", fontsize=16)
        plt.subplots_adjust(top=0.95)
        return g.figure


problem_type_map = {
    "classification": Classification,
    "object_detection": ObjectDetection,
    "semantic_segmentation": SemanticSegmentation,
}


@metric_manager.register("semantic_segmentation.dice_distribution")
def calculate_dice_distribution(data: dict, problem_type: str) -> dict:
    """
    A function to calculate the dice distribution for semantic segmentation

    :parm data: a dictionary containing the ground truth and prediction data
    :type data: dict
    :param problem_type: the type of problem
    :type problem_type: str

    :return: a dictionary containing the dice distribution
    :rtype: dict
    """
    metric_manager = problem_type_map[problem_type]()
    result = metric_manager.calculate(data)
    return result


@plot_manager.register("semantic_segmentation.dice_distribution")
def plot_dice_distribution(
    results: dict,
    save_plot: bool,
    file_name: Optional[str] = "dice_distribution.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the threshold

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save the plot
    :type save_plot: bool
    :param file_name: the name of the file
    :type file_name: str
    :param cohort_id: the cohort id
    :type cohort_id: int

    :return: the file path
    :rtype: str
    """
    plotter = PlotDiceDistribution(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()


from ._managers.metric_manager import metric_manager
from ._managers.plot_manager import plot_manager
from ._metrics.common import (
    auc,
    confusion_matrix,
    lift_chart,
    most_confused,
    stats_tables,
    top_losses,
    threshold,
    average_precision,
    predicted_distribution,
    confidence_distribution,
    iou_distribution,
    dice_distribution,
)



from typing import Union, Optional
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager
from .iou import IoUCalc


class Classification:
    pass


class SemanticSegmentation:
    pass


class ObjectDetection:
    def __init__(self):
        self.iou = IoUCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        """
        # check for the important keys in the data
        check_keys = ("ground_truth", "prediction")
        for _key in check_keys:
            if _key not in data:
                raise ValueError(f"Missing {_key} in the data dictionary")

        # check the common set of images
        common_ids = set(list(data["prediction"].keys())).difference(
            set(list(data["ground_truth"].keys()))
        )

        if common_ids:
            raise ValueError(
                "prediction and ground truth does not have corresponding samples"
            )

    def _preprocess(self, data: dict) -> tuple:
        """
        A function to preprocess

        :param data: dictionary data
        :type data: dict

        :return: gt, pred
        :rtype: tuple(dict, dict)
        """
        from .average_precision import ObjectDetection

        return ObjectDetection._preprocess(data, get_label=False, get_pred_scores=True)

    def _calculate_conf_dist(
        self, gt_boxes_dict, pred_boxes_dict: dict
    ) -> pd.DataFrame:
        """
        A function to organize the data for plotting the confidence distribution

        :param gt_boxes_dict: a dictionary containing the ground truth boxes
        :type gt_boxes_dict: dict
        :param pred_boxes_dict: a dictionary containing the prediction boxes
        :type pred_boxes_dict: dict

        :return: a pandas dataframe containing the organized data
        :rtype: pd.DataFrame

        """
        results = []
        for image_id, pred_boxes in pred_boxes_dict.items():
            gt_boxes = gt_boxes_dict[image_id]
            matched_gt = [False] * len(gt_boxes)

            for pred_box in pred_boxes:
                score = pred_box[-1]
                pred_box = pred_box[:-1]
                best_iou = 0
                best_gt_idx = -1

                for idx, gt_box in enumerate(gt_boxes):
                    iou = self.iou.calculate(pred_box, gt_box)
                    if iou > best_iou:
                        best_iou = iou
                        best_gt_idx = idx

                if best_iou > 0.5:
                    if not matched_gt[best_gt_idx]:
                        results.append(
                            {"confidence": score, "label": "TP", "best_iou": best_iou}
                        )
                        matched_gt[best_gt_idx] = True
                    else:
                        results.append(
                            {"confidence": score, "label": "FP", "best_iou": best_iou}
                        )
                else:
                    results.append(
                        {"confidence": score, "label": "FP", "best_iou": best_iou}
                    )

        data = pd.DataFrame(results)
        return data

    def _calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics for object detection

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        :param class_mapping: a dictionary containing the class mapping
        :type class_mapping: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        results = {}

        # preprocess the data
        gt_boxes_dict, pred_boxes_dict = self._preprocess(data)

        # re-organize the confidence distribution
        result = self._calculate_conf_dist(gt_boxes_dict, pred_boxes_dict)
        results["confidence_distribution"] = result

        return results

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the metrics for object detection

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        result = {}

        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self._calculate_metrics(data, data.get("class_mapping"))
        return result


class PlotConfidenceDistribution:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        validates the data required for plotting the bar plot
        """
        if "confidence_distribution" not in self.data:
            raise ValueError("confidence_distribution data is missing.")

        if not isinstance(self.data["confidence_distribution"], pd.DataFrame):
            raise ValueError(f"Data must be a data frame.")

    def save(self, fig: Figure, filename: str) -> str:
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        A function to plot the confidence distribution
        """
        sns.set_theme(style="whitegrid")
        self._validate_data()

        plot_data = self.data["confidence_distribution"]

        g = sns.JointGrid(
            data=plot_data,
            x="confidence",
            y="best_iou",
            hue="label",
            space=0,
            height=8,
            ratio=7,
        )

        g.plot_joint(sns.scatterplot, palette="pastel")
        g.plot_marginals(sns.histplot, kde=True, color=".5")

        g.set_axis_labels("Confidence Score", "Best IoU", fontsize=14)
        g.figure.suptitle(
            "Scatterplot with confidence distribution histograms", fontsize=16
        )
        plt.subplots_adjust(top=0.95)

        return g.figure


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("object_detection.confidence_distribution")
def calculate_confidence_distribution(data: dict, problem_type: str):
    """
    A wrapper function to calculate the confidence_distribution metrics.

    :param data: a dictionary containing the ground truth and prediction data
    :type data: dict
    :param problem_type: the type of problem
    :type problem_type: str

    :return: a dictionary containing the calculated metrics
    :rtype: dict
    """
    _metric_calculator = problem_type_map[problem_type]()
    result = _metric_calculator.calculate(data)
    return result


@plot_manager.register("object_detection.confidence_distribution")
def plot_confidence_distribution_od(
    results: dict,
    save_plot: bool,
    file_name: str = "confidence_distribution.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the confidence distribution metrics.

    :param results: a dictionary containing the calculated metrics
    :type results: dict
    :param save_plot: a flag to save the plot
    :type save_plot: bool
    :param file_name: the name of the file to save the plot
    :type file_name: str
    :param cohort_id: the cohort id
    :type cohort_id: int

    :return: the file path of the saved plot
    :rtype: str
    """
    plotter = PlotConfidenceDistribution(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



from typing import Union, Optional
import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager
from .iou import IoUCalc
from .dice import DiceCalc


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(
        self, data: dict, get_logits: bool = False, keep_imageid: bool = False
    ):
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean
        :param keep_imageid: to keep the image id in the response set to True
        :type keep_imageid: bool

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth, image_id_list = [], [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])
            image_id_list.append(image_id)

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])

        return (np.asarray(prediction), np.asarray(ground_truth), image_id_list)

    @staticmethod
    def _softmax(logits) -> np.ndarray:
        """
        A function to calculate the softmax

        :param logits: logits
        :type logits: np.ndarray

        :return: computed probabilities from logits
        :rtype: np.ndarray
        """
        exp_logits = np.exp(logits - np.max(logits, axis=1, keepdims=True))
        probabilities = exp_logits / np.sum(exp_logits, axis=1, keepdims=True)
        return probabilities

    def _calculate_loss(
        self, y_true: np.ndarray, y_pred: np.ndarray, overall: bool = False
    ) -> float:
        # Convert logits to probabilities
        probabilities = self._softmax(y_pred)
        # Clip probabilities to prevent log(0)
        probabilities = np.clip(probabilities, 1e-15, 1 - 1e-15)
        # Create one-hot encoded labels
        y_true_one_hot = np.eye(np.max(y_true) + 1)[y_true]
        # Calculate cross-entropy loss

        if overall:
            # to calculate if the loss is calculated over all the examples
            loss = -np.sum(y_true_one_hot * np.log(probabilities)) / y_true.shape[0]
        else:
            loss = -np.log(probabilities[np.arange(len(y_true)), y_true])

        return np.round(loss, 4)

    def __calculate_metrics(self, data: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        prediction, ground_truth, image_id = self._preprocess(
            data, get_logits=True, keep_imageid=True
        )
        pred_gt_df = pd.DataFrame(prediction)
        pred_gt_df["ground_truth"] = ground_truth.tolist()
        pred_gt_df["image_id"] = image_id

        # calculate the loss first
        # the loss considered is cross entropy loss as its the most general one used
        # for classification. Inorder to use a different loss function new class
        # method is supposed to defined and following line is replaced
        pred_gt_df["loss"] = self._calculate_loss(
            pred_gt_df["ground_truth"].to_numpy(), pred_gt_df[[0, 1]].to_numpy()
        )
        overall_loss = self._calculate_loss(
            pred_gt_df["ground_truth"].to_numpy(),
            pred_gt_df[[0, 1]].to_numpy(),
            overall=True,
        )

        # only the first 10 values with loss are picked up from the results
        # where the loss are sorted from highest to lowest order
        pred_gt_df = pred_gt_df.sort_values(by="loss", ascending=False)

        result = {"loss_data": pred_gt_df, "overall_loss": overall_loss}
        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the top losses for the given dataset.

        :param data: The input data required for calculation and plotting
                {"prediction":, "ground_truth": , "metadata":, "class_mappings":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self.__calculate_metrics(data)
        return result


class SemanticSegmentation:
    def __init__(self):
        self.dice = DiceCalc()

    def _validate_data(self, data: dict) -> bool:
        return ObjectDetection._validate_data(ObjectDetection, data)

    def _preprocess(self, data: dict) -> tuple:
        """
        function to preprocess the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict

        :return: data tuple
        :rtype: tuple
        """
        from .iou_distribution import SemanticSegmentation

        return SemanticSegmentation._preprocess(SemanticSegmentation, data)

    def _calculate_metrics(self, data: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the calculated metrics
        :rtype: dict
        """
        # preprocess the data
        gt_data, pred_data = self._preprocess(data)

        # calculate the loss
        result = {"image_id": [], "loss": []}

        for image_id in gt_data:
            gt_masks = gt_data[image_id]
            if not gt_masks:
                continue

            for gt_mask in gt_masks:
                for pred_mask in pred_data[image_id]:
                    dice_loss = self.dice.dice_loss(gt_mask, pred_mask)
                    result["image_id"].append(image_id)
                    result["loss"].append(dice_loss)

            result = pd.DataFrame(result)
            return result

    def calculate(self, data: dict) -> dict:
        """
        A function to calculate the dice loss

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict

        :return: a dictionary containing the dice distribution
        :rtype: dict
        """
        result = {}

        # validate the data
        self._validate_data(data)

        # calculate the metrics
        result["loss_data"] = self._calculate_metrics(data)
        result["overall_loss"] = result["loss_data"]["loss"].mean()
        return result


class ObjectDetection:
    def __init__(self):
        self.iou = IoUCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")

        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(self, data: dict) -> tuple:
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict

        :return: data tuple
        :rtype: tuple
        """
        from .average_precision import ObjectDetection

        return ObjectDetection._preprocess(data)

    def _calculate_loss(
        self, gt_boxes_dict: dict, pred_boxes_dict: dict, class_mapping: dict
    ) -> pd.DataFrame:
        """
        A function to calculate the loss function

        :param gt_boxes_dict: a dictionary of gt boxes, with image id as the key
        :type gt_boxes_dict: dict
        :param pred_boxes_dict: a dictionary of pred boxes, with image id as the key
        :type pred_boxes_dict: dict
        :param class_mapping: dict
        :type class_mapping: dict

        :return: pandas data frame
        """
        loss_data = {"image_id": [], "loss": []}
        for image_id in gt_boxes_dict:
            gt_box = gt_boxes_dict[image_id]
            pred_box = pred_boxes_dict[image_id]

            iou_loss = self.iou.calculate_iou_loss(gt_box, pred_box)
            loss_data["image_id"].append(image_id)
            loss_data["loss"].append(iou_loss)

        loss_data = pd.DataFrame(loss_data)
        loss_data = loss_data.sort_values(by="loss", ascending=False)
        return loss_data

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        results = {}

        # preprocess the data to load the data
        gt_boxes, pred_boxes = self._preprocess(data)

        # calculate the loss
        _loss = self._calculate_loss(gt_boxes, pred_boxes, class_mapping)

        results["loss_data"] = _loss

        return results

    def calculate(self, data: dict) -> dict:
        """
        Calculates the Top Losses metric for the given dataset.

        :param data: The input data required for calculation and plotting
                     {"prediction":, "ground_truth": , "class_mappings":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate results
        result = self.__calculate_metrics(data, data.get("class_mapping"))

        result["overall_loss"] = round(result["loss_data"]["loss"].mean(), 4)

        return result


class PlotTopLosses:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting the top losses.
        """
        if "loss_data" not in self.data:
            raise ValueError("Data must contain 'loss_data'.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file.

        :param filename: Path where the plot image will be saved
        :type filename: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self, top_k: int = 9) -> Figure:
        """
        Plots the top losses.

        :param top_k: Number of top losses to display
        :type top_k: int
        :return: Matplotlib Figure object
        :rtype: Figure
        """
        # Validate the data
        self._validate_data()
        sns.set_style("whitegrid")
        fig, ax = plt.subplots(figsize=(10, 7))
        plot_data = self.data["loss_data"].head(20)
        hue_value = "ground_truth" if "ground_truth" in plot_data.columns else None
        sns.barplot(
            data=plot_data,
            x="image_id",
            y="loss",
            ax=ax,
            palette="pastel",
            hue=hue_value,
        )

        ax.set_xlabel("Image id", fontdict={"fontsize": 14, "fontweight": "medium"})
        ax.set_xticklabels(ax.get_xticklabels(), rotation=45, ha="right")
        ax.set_ylabel("Loss Value", fontdict={"fontsize": 14, "fontweight": "medium"})

        title_txt = f"Top losses : Overall loss: {self.data['overall_loss']} : Top 20"
        if self.cohort_id:
            title_str = f"{title_txt}: cohort - {self.cohort_id}"
        else:
            title_str = f"{title_txt}"

        ax.set_title(title_str, fontdict={"fontsize": 16, "fontweight": "medium"})
        ax.legend(loc="lower right")
        return fig


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("semantic_segmentation.top_losses")
@metric_manager.register("object_detection.top_losses")
@metric_manager.register("classification.top_losses")
def calculate_top_losses_classification(data: dict, problem_type: str):
    """
    A wrapper function to calculate the top losses metric.

    :param data: Dictionary of data: {"prediction": , "ground_truth": , "loss": }
    :type data: dict
    :param problem_type: Type of the problem
    :type problem_type: str

    :return: Calculated results
    :rtype: dict
    """
    metric_calculator = problem_type_map[problem_type]()
    result = metric_calculator.calculate(data)
    return result


@plot_manager.register("semantic_segmentation.top_losses")
@plot_manager.register("object_detection.top_losses")
@plot_manager.register("classification.top_losses")
def plot_top_losses_classification(
    results: dict,
    save_plot: bool,
    file_name: str = "top_losses.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the top losses chart.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plot
    :type save_plot: bool
    :param file_name: Name of the file to save the plot
    :type file_name: str
    :param cohort_id: id of the cohort
    :type cohort_id: int

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotTopLosses(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



import os
import json
import bson
from typing import Union, Optional

import pandas as pd

from gesund.core.schema import UserInputParams, UserInputData
from gesund.core._exceptions import MetricCalculationError
from gesund.core._data_loaders import DataLoader
from gesund.core._converters import ConverterFactory
from ._result import ValidationResult
from gesund.core._managers.metric_manager import metric_manager


def categorize_age(age):
    if age < 18:
        return "Child"
    elif 18 <= age < 30:
        return "Young Adult"
    elif 30 <= age < 60:
        return "Adult"
    else:
        return "Senior"


class Validation:
    def __init__(
        self,
        annotations_path: str,
        predictions_path: str,
        class_mapping: Union[str, dict],
        problem_type: str,
        data_format: str,
        json_structure_type: str,
        plot_config: str,
        metadata_path: Optional[str] = None,
        cohort_args: Optional[dict] = {},
        plot_args: Optional[dict] = {},
        metric_args: Optional[dict] = {},
    ):
        """
        Initialization function to handle the validation pipeline

        :param annotations_path: Path to the JSON file containing the annotations data.
        :type annotations_path: str
        :param predictions_path: Path to the JSON file containing the predictions data.
        :type predictions_path: str
        :param class_mappings: Path to the JSON file containing class mappings or a dictionary file.
        :type class_mappings: Union[str, dict]
        :param problem_type: Type of problem (e.g., 'classification', 'object_detection').
        :type problem_type: str
        :param json_structure_type: Data format for the validation (e.g., 'coco', 'yolo', 'gesund').
        :type json_structure_type: str
        :param plot_config: Config for the plotting
        :type plot_config: dict
        :param metadata_path: Path to the metadata file (if available).
        :type metadata_path: str
        :optional metadata_path: true
        :param cohort_args: arguments supplied for cohort analysis
        :type cohort_args: dict
        :param plot_args: arguments supplied for cohort analysis
        :type plot_args: dict
        :param metric_args: arguments supplied for validation metrics
        :type metric_args: dict

        :return: None
        """
        # set up user parameters
        params = {
            "annotations_path": annotations_path,
            "predictions_path": predictions_path,
            "metadata_path": metadata_path,
            "problem_type": problem_type,
            "json_structure_type": json_structure_type,
            "data_format": data_format,
            "class_mapping": class_mapping,
            "plot_config": plot_config,
        }
        self.user_params = UserInputParams(**params)

        # set up and load all the required data
        self._load_data()

        # set up batch job id
        self.batch_job_id = str(bson.ObjectId())
        self.output_dir = f"outputs/{self.batch_job_id}"

        os.makedirs(self.output_dir, exist_ok=True)
        os.makedirs(os.path.join(self.output_dir, "plots"), exist_ok=True)

        # cohort parameters
        self.cohort_params = {
            "max_num_cohort": 10,
            "min_cohort_size": 0,
            "cohort_map": {},
        }
        if cohort_args:
            self.cohort_params.update(cohort_args)

        self.plot_args = plot_args
        self.metric_args = metric_args

        self.debug_mode = False

    def _load_data(self) -> dict:
        """
        A Function to load the JSON files

        :return: None
        """
        # Load data
        # set up source data for processing
        data_loader = DataLoader(self.user_params.data_format)
        data = {
            "prediction": data_loader.load(self.user_params.predictions_path),
            "annotation": data_loader.load(self.user_params.annotations_path),
        }
        if isinstance(self.user_params.class_mapping, str):
            data["class_mapping"] = data_loader.load(self.user_params.class_mapping)
        else:
            data["class_mapping"] = self.user_params.class_mapping

        if self.user_params.metadata_path:
            data["metadata"] = data_loader.load(self.user_params.metadata_path)

        # run conversion
        if self.user_params.json_structure_type != "gesund":
            self._convert_data(data)

        self.data = UserInputData(**data)

    def apply_metadata(self, prediction: dict, annotation: dict) -> dict:
        """
        A function to create cohorts as per the metadata

        :param prediction: structure containing the prediction data
        :type prediction: dict
        :param annotation: structure containing the annotation data
        :type annotation: dict

        :return: Filtered dataset
        :rtype: dict
        """
        # convert the metadata into pandas dataframe for better access
        df: pd.DataFrame = pd.DataFrame.from_records(self.data.metadata)
        cohorts_data = {}
        cohort_id_map = {}

        # collect only the metadata columns
        metadata_columns = df.columns.tolist()
        metadata_columns.remove("image_id")
        lower_case = {i: i.lower() for i in metadata_columns}
        df = df.rename(columns=lower_case)

        # categorize age in metadata
        if "age" in list(lower_case.values()):
            df["age"] = df["age"].apply(categorize_age)

        # loop over group by to form cohorts with unique metadata
        cohort_id = 0
        for grp, subset_data in df.groupby(list(lower_case.values())):
            grp_str = ",".join([str(i) for i in grp])
            cohort_id = cohort_id + 1
            cohort_id_map[cohort_id] = {"group": grp_str, "size": subset_data.shape[0]}

            # it seems that
            if subset_data.shape[0] < self.cohort_params["min_cohort_size"]:
                print(
                    f"Warning - grp excluded - {grp_str} cohort size < {self.cohort_params['min_cohort_size']}"
                )
            else:
                image_ids = set(subset_data["image_id"].to_list())
                filtered_data = {
                    "prediction": {
                        i: prediction[i] for i in prediction if i in image_ids
                    },
                    "ground_truth": {
                        i: annotation[i] for i in annotation if i in image_ids
                    },
                    "metadata": subset_data,
                    "class_mapping": self.data.class_mapping,
                }
                cohorts_data[cohort_id] = filtered_data

        self.cohort_params["cohort_map"] = cohort_id_map
        return cohorts_data

    def _convert_data(self, data):
        """
        A function to convert the data from the respective structure to the gesund format

        :param data: dictionary containing the data
        :type data: dict

        :return: data dictionary
        :rtype: dict
        """
        # setup data converter
        data_converter = ConverterFactory().get_converter(
            self.user_params.json_structure_type
        )

        # run the converters
        (
            data["converted_annotation"],
            data["converted_prediction"],
        ) = data_converter.convert(
            annotation=data["annotation"],
            prediction=data["prediction"],
            problem_type=self.user_params.problem_type,
        )
        data["was_converted"] = True
        return data

    def _run_validation(self, data: dict) -> dict:
        """
        A function to run the validation

        :param: None

        :return: result dictionary
        :rtype: dict
        """
        results = {}
        try:
            for metric_name in metric_manager.get_names(
                problem_type=self.user_params.problem_type
            ):
                key_name = f"{self.user_params.problem_type}.{metric_name}"
                print("Running ", key_name, "." * 5)
                _metric_executor = metric_manager[key_name]
                _result = _metric_executor(
                    data=data, problem_type=self.user_params.problem_type
                )
                results[metric_name] = _result
        except Exception as e:
            print(e)
            raise MetricCalculationError("Error in calculating metrics!")

        return results

    def _save_json(self, results) -> None:
        """
        A function to save the validation results

        :param results: Dictionary containing the results
        :type results: dict

        :return: None
        """
        for plot_name, metrics in results.items():
            output_file = os.path.join(self.output_dir, f"{plot_name}.json")
            try:
                with open(output_file, "w") as f:
                    json.dump(metrics, f, indent=4)
            except Exception as e:
                print(f"Could not save metrics for {plot_name} because: {e}")

    @staticmethod
    def format_metrics(metrics: dict) -> dict:
        """
        Format and print the overall metrics in a readable format.

        This function takes in the metrics data, formats it, and prints out the
        highlighted overall metrics, including confidence intervals when applicable.
        It also prints a message indicating that all graphs and plot metrics have been saved.

        :param metrics: (dict) A dictionary containing the metrics, expected to have
                        a 'plot_highlighted_overall_metrics' key with the metrics data.

        :return: None
        """

        print("\nOverall Highlighted Metrics:\n" + "-" * 40)
        for metric, values in metrics["plot_highlighted_overall_metrics"][
            "data"
        ].items():
            print(f"{metric}:")
            for key, value in values.items():
                if isinstance(value, list):  # If it's a confidence interval
                    value_str = f"{value[0]:.4f} to {value[1]:.4f}"
                else:
                    value_str = f"{value:.4f}"
                print(f"    {key}: {value_str}")
            print("-" * 40)
        print("All Graphs and Plots Metrics saved in JSONs.\n" + "-" * 40)

    def run(self) -> ValidationResult:
        """
        A function to run the validation pipeline

        :param: None
        :type:

        :return: None
        :rtype:
        """
        results = {}
        cohorts = None

        # the data is assigned to local veriables to later pass on to functions as parameters
        # eventhough the data is stored in class attribute, for better code clarity
        if self.data.was_converted:
            prediction = self.data.converted_prediction
            annotation = self.data.converted_annotation
        else:
            prediction = self.data.prediction
            annotation = self.data.annotation

        # if the metadata is made available then the data is divided into cohorts as per the metadata
        if self.user_params.metadata_path:
            cohorts = self.apply_metadata(prediction, annotation)

        # run the validation
        if cohorts:
            for _cohort_id in cohorts:
                data = cohorts[_cohort_id]
                data["metric_args"] = self.metric_args
                results[_cohort_id] = self._run_validation(data)
        else:
            results = self._run_validation(
                data={
                    "prediction": prediction,
                    "ground_truth": annotation,
                    "metadata": self.data.metadata,
                    "class_mapping": self.data.class_mapping,
                    "metric_args": self.metric_args,
                }
            )

        # return the results
        return ValidationResult(
            data=self.data,
            input_params=self.user_params,
            result=results,
            plot_args=self.plot_args,
            cohort_args=self.cohort_params,
        )



from setuptools import setup

setup(
    packages=["gesund-src"],
    install_requires=[
        "bson",
        "jsonschema",
        "scikit-learn",
        "pandas",
        "seaborn",
        "opencv-python",
        "dictances==1.5.3",
        "miseval==1.2.2",
        "numpy==1.21.6",
        "numba==0.55.2",
        "tqdm",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    include_package_data=True,
)



import os
from typing import Union, Dict, List, Optional

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import seaborn as sns

from gesund.core import metric_manager, plot_manager


class Classification:
    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the data that is required for metric calculation and plotting.

        :param data: The input data required for calculation, {"prediction":, "ground_truth": , "metadata":}
        :type data: dict

        :return: Status if the data is valid
        :rtype: bool
        """
        # Basic validation checks
        if not isinstance(data, dict):
            raise ValueError("Data must be a dictionary.")
        required_keys = ["prediction", "ground_truth"]
        for key in required_keys:
            if key not in data:
                raise ValueError(f"Data must contain '{key}'.")

        if len(data["prediction"]) != len(data["ground_truth"]):
            raise ValueError("Prediction and ground_truth must have the same length.")

        if (
            len(
                set(list(data["prediction"].keys())).difference(
                    set(list(data["ground_truth"].keys()))
                )
            )
            > 0
        ):
            raise ValueError("Prediction and ground_truth must have the same keys.")

        return True

    def _preprocess(self, data: dict, get_logits=False):
        """
        Preprocesses the data

        :param data: dictionary containing the data prediction, ground truth, metadata
        :type data: dict
        :param get_logits: in case of multi class classification set to True
        :type get_logits: boolean

        :return: data tuple
        :rtype: tuple
        """
        prediction, ground_truth = [], []
        for image_id in data["ground_truth"]:
            sample_gt = data["ground_truth"][image_id]
            sample_pred = data["prediction"][image_id]
            ground_truth.append(sample_gt["annotation"][0]["label"])

            if get_logits:
                prediction.append(sample_pred["logits"])
            else:
                prediction.append(sample_pred["prediction_class"])
        return (np.asarray(prediction), np.asarray(ground_truth))

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to calculate the metrics

        :param data: data dictionary containing data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        class_order = [int(i) for i in class_mapping.keys()]

        prediction, ground_truth = self._preprocess(data, get_logits=True)

        lift_chart_calculator = LiftChart(class_mapping)
        lift_points = lift_chart_calculator.calculate_lift_curve_points(
            ground_truth, prediction
        )

        result = {
            "lift_points": lift_points,
            "class_mapping": class_mapping,
            "class_order": class_order,
        }

        return result

    def calculate(self, data: dict) -> dict:
        """
        Calculates the lift chart for the given data.

        :param data: The input data required for calculation and plotting
                     {"prediction":, "ground_truth": , "metadata":, "class_mapping":}
        :type data: dict

        :return: Calculated metric results
        :rtype: dict
        """
        result = {}

        # Validate the data
        self._validate_data(data)

        # calculate the metrics
        result = self.__calculate_metrics(data, data.get("class_mapping"))
        return result


class LiftChart:
    def __init__(self, class_mapping: Dict[int, str]) -> None:
        self.class_mapping = class_mapping
        self.class_order = [int(i) for i in list(class_mapping.keys())]

    def _decile_table(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_logits: pd.DataFrame,
        predicted_class: int = 0,
        change_deciles: int = 20,
        labels: bool = True,
        round_decimal: int = 3,
    ) -> pd.DataFrame:
        """
        Generates the Decile Table from labels and probabilities.

        :return: Dataframe containing decile metrics
        """
        df = pd.DataFrame(
            {"true": true, "pred_logits": pred_logits.loc[:, predicted_class]}
        )

        # Sort by predicted logits in descending order
        df.sort_values("pred_logits", ascending=False, inplace=True)

        # Assign deciles
        df["decile"] = pd.qcut(
            df["pred_logits"].rank(method="first"), change_deciles, labels=False
        )
        df["decile"] = df["decile"] + 1  # Deciles start from 1

        # Calculate lift metrics
        lift_df = (
            df.groupby("decile")
            .agg(
                prob_min=("pred_logits", "min"),
                prob_max=("pred_logits", "max"),
                prob_avg=("pred_logits", "mean"),
                cnt_cust=("pred_logits", "count"),
                cnt_resp=("true", "sum"),
            )
            .reset_index()
        )

        lift_df["cnt_non_resp"] = lift_df["cnt_cust"] - lift_df["cnt_resp"]

        total_resp = df["true"].sum()
        total_cust = df["true"].count()

        lift_df["resp_rate"] = round(
            lift_df["cnt_resp"] * 100 / lift_df["cnt_cust"], round_decimal
        )
        lift_df["cum_cust"] = np.cumsum(lift_df["cnt_cust"])
        lift_df["cum_resp"] = np.cumsum(lift_df["cnt_resp"])
        lift_df["cum_non_resp"] = np.cumsum(lift_df["cnt_non_resp"])
        lift_df["cum_cust_pct"] = round(
            lift_df["cum_cust"] * 100 / total_cust, round_decimal
        )
        lift_df["cum_resp_pct"] = round(
            lift_df["cum_resp"] * 100 / total_resp, round_decimal
        )
        lift_df["cum_non_resp_pct"] = round(
            lift_df["cum_non_resp"] * 100 / (total_cust - total_resp), round_decimal
        )
        lift_df["KS"] = round(
            lift_df["cum_resp_pct"] - lift_df["cum_non_resp_pct"], round_decimal
        )
        lift_df["lift"] = round(
            lift_df["cum_resp_pct"] / lift_df["cum_cust_pct"], round_decimal
        )
        return lift_df

    def calculate_lift_curve_points(
        self,
        true: Union[List[int], np.ndarray, pd.Series],
        pred_logits: Union[List[int], np.ndarray, pd.Series],
        predicted_class: Optional[int] = None,
        change_deciles: int = 20,
        labels: bool = True,
        round_decimal: int = 3,
    ) -> Dict[str, List[Dict[str, float]]]:
        pred_df = pd.DataFrame(pred_logits)
        class_lift_dict = dict()
        if predicted_class in [None, "all", "overall"]:
            for class_ in self.class_order:
                lift_df = self._decile_table(
                    true,
                    pred_df,
                    predicted_class=int(class_),
                    change_deciles=change_deciles,
                    labels=labels,
                    round_decimal=round_decimal,
                )
                xy_points = [
                    {"x": avg_prob, "y": lift}
                    for avg_prob, lift in zip(lift_df["prob_avg"], lift_df["lift"])
                ]
                class_name = self.class_mapping[str(class_)]
                class_lift_dict[class_name] = xy_points
        else:
            lift_df = self._decile_table(
                true,
                pred_df,
                predicted_class=predicted_class,
                change_deciles=change_deciles,
                labels=labels,
                round_decimal=round_decimal,
            )
            xy_points = [
                {"x": avg_prob, "y": lift}
                for avg_prob, lift in zip(lift_df["prob_avg"], lift_df["lift"])
            ]
            class_name = self.class_mapping[str(predicted_class)]
            class_lift_dict[class_name] = xy_points

        return class_lift_dict


class SemanticSegmentation(Classification):
    pass


class ObjectDetection:
    pass


class PlotLiftChart:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.lift_points = data["lift_points"]
        self.class_mapping = data["class_mapping"]
        self.class_order = data["class_order"]
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        Validates the data required for plotting the Lift Chart.
        """
        if "lift_points" not in self.data:
            raise ValueError("Data must contain 'lift_points'.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file.

        :param filename: Path where the plot image will be saved
        :type filename: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        Plots the Lift Chart.

        :return: Matplotlib Figure object
        :rtype: Figure
        """
        sns.set_style("whitegrid")
        # Validate the data
        self._validate_data()

        fig, ax = plt.subplots(figsize=(10, 7))
        for class_name, points in self.lift_points.items():
            x = [p["x"] for p in points]
            y = [p["y"] for p in points]
            ax.plot(x, y, marker="o", label=f"Class {class_name}")

        ax.set_xlabel(
            "Average Predicted Probability",
            fontdict={"fontsize": 14, "fontweight": "medium"},
        )
        ax.set_ylabel(
            "Calculated Lift", fontdict={"fontsize": 14, "fontweight": "medium"}
        )

        if self.cohort_id:
            title_str = f"Lift Chart : cohort - {self.cohort_id}"
        else:
            title_str = "Lift Chart"

        ax.set_title(title_str, fontdict={"fontsize": 16, "fontweight": "medium"})
        ax.legend()
        return fig


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("classification.lift_chart")
def calculate_lift_chart_metric(data: dict, problem_type: str):
    """
    A wrapper function to calculate the Lift Chart metric.

    :param data: Dictionary of data: {"prediction": , "ground_truth": }
    :type data: dict
    :param problem_type: Type of the problem
    :type problem_type: str

    :return: Dict of calculated results
    :rtype: dict
    """
    _metric_calculator = problem_type_map[problem_type]()
    result = _metric_calculator.calculate(data)
    return result


@plot_manager.register("classification.lift_chart")
def plot_lift_chart(
    results: dict,
    save_plot: bool,
    file_name: str = "lift_chart.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot the lift chart.

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: Boolean value to save plot
    :type save_plot: bool
    :param file_name: Name of the file to save the plot
    :type file_name: str
    :param cohort_id: id of the cohort
    :type cohort_id: int

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotLiftChart(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()



from typing import Union, Dict, List, Optional
import numpy as np


def mask_to_rle(mask: np.ndarray) -> List[int]:
    """Convert a 2D binary mask to RLE format."""
    try:
        pixels = mask.flatten()
        runs = np.where(pixels[1:] != pixels[:-1])[0] + 1
        runs = np.concatenate([[0], runs, [len(pixels)]])
        rle = []
        for start, stop in zip(runs[:-1], runs[1:]):
            if pixels[start] == 1:
                rle.extend([start, stop - start])
        return rle
    except Exception as e:
        print(e)


class ClassificationConverter:
    def __init__(
        self,
        annotations: list,
        predictions: list,
        image_width: int = 512,
        image_height: int = 512,
    ):
        """
        The initialization function of the class

        :param annotations: list of annotations
        :param predictions: list of predictions
        :param image_width: integer value representing the image width
        :param image_height: integer value representing the image height
        :return: None
        """
        self.annotations = annotations
        self.predictions = predictions
        self.image_width = image_width
        self.image_height = image_height

    def _convert_predictions(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a dictionary of predictions in the gesund predictions format
        """
        custom_json = {}

        for item in self.predictions:
            image_id = item["image_id"]

            for prediction in item["predictions"]:
                class_id = prediction["class"]
                confidence = prediction["confidence"]
                loss = prediction["loss"]
                logits = [0.0, 0.0]
                logits[class_id] = confidence
                logits[1 - class_id] = 1 - confidence

                pred = {
                    "image_id": image_id,
                    "prediction_class": class_id,
                    "confidence": confidence,
                    "logits": logits,
                    "loss": loss,
                }

                # if image_id in custom_json:
                #     custom_json[image_id].append(pred)
                # else:
                custom_json[image_id] = pred
        return custom_json

    def _convert_annotations(self) -> dict:
        """
        A function to convert the yolo annotations to gesund annotations format

        :return: a dictionary of annotations in the gesund annotations format
        """
        custom_json = {}
        for item in self.annotations:
            image_id = item["image_id"]

            for annotation in item["annotations"]:
                class_id = annotation["class"]

                if image_id in custom_json:
                    custom_json[image_id]["annotation"].append({"label": class_id})
                else:
                    custom_json[image_id] = {
                        "image_id": image_id,
                        "annotation": [
                            {
                                "label": class_id,
                            }
                        ],
                    }
        return custom_json


class ObjectDetectionConverter:
    def __init__(
        self,
        annotations: Union[list, dict],
        predictions: Union[list, dict],
        image_width: int,
        image_height: int,
    ):
        """
        The initialization function of the class


        :param annotations: a list of annotation in the yolo format to convert to gesund format
        :type annotations: list
        :param predictions: a list of predictions in the yolo format to convert into the gesund format
        :type predictions: list
        :param image_width: The width of the image
        :type image_width: int
        :param image_height: The height of the image
        :type image_height: int

        :return: None
        """
        self.annotation = annotations
        self.predictions = predictions
        self.image_width = image_width
        self.image_height = image_height

    def _convert_predictions(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a dictionary of predictions in the gesund predictions format
        :rtype: dict
        """
        custom_json = {}

        for item in self.predictions:
            image_id = item["image_id"]
            annotations = item["annotations"]

            for annotation in annotations:
                class_id = annotation["class"]
                x_center = annotation["x_center"]
                y_center = annotation["y_center"]
                width = annotation["width"]
                height = annotation["height"]

                # convert normalized to pixel values
                x1 = int((x_center - width / 2) * self.image_width)
                y1 = int((y_center - height / 2) * self.image_height)
                x2 = int((x_center + width / 2) * self.image_width)
                y2 = int((y_center + height / 2) * self.image_height)

                custom_prediction = {
                    "box": {"x1": x1, "y1": y1, "x2": x2, "y2": y2},
                    "confidence": annotation["confidence"],
                    "prediction_class": class_id,
                }

                if image_id in custom_json:
                    custom_json[image_id]["objects"].append(custom_prediction)
                else:
                    custom_json[image_id] = {
                        "objects": [],
                        "shape": [self.image_width, self.image_height],
                        "status": 200,
                        "image_id": image_id,
                    }
                    custom_json[image_id]["objects"].append(custom_prediction)
        return custom_json

    def _convert_annotations(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a dictionary of predictions in the gesund predictions format
        :rtype: dict
        """
        custom_json = {}

        for item in self.annotation:
            image_id = item["image_id"]
            annotations = item["annotations"]

            custom_annotations = []
            for annotation in annotations:
                class_id = annotation["class"]
                x_center = annotation["x_center"]
                y_center = annotation["y_center"]
                width = annotation["width"]
                height = annotation["height"]

                # convert normalized values to pixel values
                x1 = int((x_center - width / 2) * self.image_width)
                y1 = int((y_center - height / 2) * self.image_height)
                x2 = int((x_center + width / 2) * self.image_width)
                y2 = int((y_center + height / 2) * self.image_height)

                custom_annotation = {
                    "label": f"class_{class_id}",
                    "points": [{"x": x1, "y": y1}, {"x": x2, "y": y2}],
                    "type": "rect",
                }
                custom_annotations.append(custom_annotation)

            custom_json[image_id] = {
                "image_id": image_id,
                "annotation": custom_annotations,
            }
        return custom_json


class SemanticSegmentationConverter:
    def __init__(
        self,
        annotations: Union[list, dict],
        predictions: Union[list, dict],
        image_width: int,
        image_height: int,
    ):
        """
        the initialization function of the class

        :param annotations: a list of annotation in the yolo format to convert to gesund format
        :type annotations: Union[list, dict]
        :param predictions: a list of predictions in the yolo format to convert into the gesund format
        :type predictions: Union[list, dict]
        :param image_width: The width of the image
        :type image_width: int
        :param image_height: The height of the image
        :type image_height: int

        :return: None
        """
        self.annotation = annotations
        self.predictions = predictions
        self.image_width = image_width
        self.image_height = image_height

    def _convert_predictions(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a list of objects in the gesund predictions format
        :rtype: dict
        """
        custom_json = {}

        for item in self.predictions:
            image_id = item["image_id"]
            predictions = item["predictions"]
            for pred in predictions:
                class_id = pred["class"]
                segmentation = pred["segmentation"]

                # convert to pixel values
                pixel_values = [
                    {
                        "x": int(val["x"] * self.image_width),
                        "y": int(val["y"] * self.image_height),
                    }
                    for val in segmentation
                ]

                # convert the pixel values to binary masks
                binary_mask = np.zeros(
                    (self.image_height, self.image_width), dtype=np.uint8
                )
                for point in pixel_values:
                    x, y = point["x"], point["y"]
                    if 0 <= x < self.image_width and 0 <= y < self.image_height:
                        binary_mask[y, x] = 1

                # convert the binary masks to RLE masks
                rle = mask_to_rle(binary_mask)
                rle = " ".join([str(i) for i in rle])
                rle = {"rle": rle, "class": class_id}

                # save the created items in the json format
                if image_id in custom_json:
                    custom_json[image_id]["masks"]["rles"].append(rle)
                else:
                    custom_json[image_id] = {
                        "image_id": image_id,
                        "masks": {"rles": [rle]},
                        "shape": [self.image_width, self.image_height],
                        "status": 200,
                    }
            return custom_json

    def _convert_annotations(self) -> dict:
        """
        A function to convert the yolo annotations to gesund predictions format

        :return: a list of objects in the gesund predictions format
        :rtype: dict
        """
        custom_json = {}

        for item in self.annotation:
            image_id = item["image_id"]
            annotations = item["annotations"]

            custom_annotations = []
            for annotation in annotations:
                class_id = annotation["class"]
                segmentation = annotation["segmentation"]

                # convert the normalized values to pixel values
                pixel_values = [
                    {
                        "x": int(val["x"] * self.image_width),
                        "y": int(val["y"] * self.image_height),
                    }
                    for val in segmentation
                ]

                # convert the pixel values to binary masks
                binary_mask = np.zeros(
                    (self.image_height, self.image_width), dtype=np.uint8
                )

                for point in pixel_values:
                    x, y = point["x"], point["y"]
                    if 0 <= x < self.image_width and 0 <= y < self.image_height:
                        binary_mask[y, x] = 1

                # convert the binary masks to RLE masks
                rle = mask_to_rle(binary_mask)
                rle = " ".join([str(i) for i in rle])
                # save the created items in the json format
                custom_annotation = {
                    "image_id": image_id,
                    "label": class_id,
                    "type": "mask",
                    "measurement_info": {
                        "objectName": "mask",
                        "measurement": "Segmentation",
                    },
                    "mask": {"mask": rle},
                    "shape": [self.image_width, self.image_height],
                    "window_level": None,
                }
                custom_annotations.append(custom_annotation)
            custom_json[image_id] = {
                "image_id": image_id,
                "annotation": custom_annotations,
            }

        return custom_json


class InstanceSegmentationConverter:
    def __init__(
        self,
        annotations: Union[list, dict],
        predictions: Union[list, dict],
        image_width: int,
        image_height: int,
    ):
        """
        The initialization function of the class


        :param annotations: a list of annotation in the yolo format to convert to gesund format
        :type annotations: Union[list, dict]
        :param predictions: a list of predictions in the yolo format to convert into the gesund format
        :type predictions: Union[list, dict]
        :param image_width: The width of the image
        :type image_width: int
        :param image_height: The height of the image
        :type image_height: int

        :return: None
        """
        self.annotation = annotations
        self.predictions = predictions
        self.image_width = image_width
        self.image_height = image_height

    def _convert_predictions(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a dict of objects in the gesund predictions format
        :rtype: dict
        """
        pass

    def _convert_annotations(self) -> dict:
        """
        A function to convert the yolo predictions to gesund predictions format

        :return: a dict of objects in the gesund predictions format
        :rtype: dict
        """
        pass


class YoloProblemTypeFactory:
    def get_yolo_converter(self, problem_type: str):
        """
        A factory method to get the yolo converter

        :param problem_type: type of the problem
        :type problem_type: str

        :return: class of the converter
        :rtype: class
        """
        if problem_type == "object_detection":
            return ObjectDetectionConverter
        elif problem_type == "semantic_segmentation":
            return SemanticSegmentationConverter
        elif problem_type == "classification":
            return ClassificationConverter
        else:
            raise NotImplementedError


class YoloToGesund:
    def convert(
        self,
        annotation: Union[List[Dict], Dict],
        prediction: Union[List[Dict], Dict],
        problem_type: str,
        image_width: Optional[int] = 512,
        image_height: Optional[int] = 512,
    ) -> Dict:
        """
        A run method to execute the pipeline and convert the given format to gesund format

        :param annotation: the annotations data
        :type annotation: Union[List[Dict], Dict]
        :param prediction: the prediction data
        :type prediction: Union[List[Dict], Dict]
        :param problem_type: problem type could be classification | object detection | segmentation
        :type problem_type: str
        :param image_width: width of the image
        :type image_width: int
        :param image_height: height of the image
        :type image_height: int

        :return: dictionary of converted format
        :rtype: dict
        """
        _converter_cls = YoloProblemTypeFactory().get_yolo_converter(
            problem_type=problem_type
        )
        _converter = _converter_cls(
            annotations=annotation,
            predictions=prediction,
            image_width=image_width,
            image_height=image_height,
        )
        return _converter._convert_annotations(), _converter._convert_predictions()



from typing import Union, Optional
import os

import numpy as np
import pandas as pd
from matplotlib.figure import Figure
import seaborn as sns
import matplotlib.pyplot as plt

from gesund.core import metric_manager, plot_manager
from .iou import IoUCalc


class AveragePrecision:
    pass


class Classification:
    pass


class SemanticSegmentation:
    pass


class ObjectDetection:
    def __init__(self):
        self.iou = IoUCalc()

    def _validate_data(self, data: dict) -> bool:
        """
        A function to validate the given data used for calculating metrics for object detection validation

        :param data: a dictionary containing the ground truth and prediction data
        :type data: dict
        """
        # check for the important keys in the data
        check_keys = ("ground_truth", "prediction", "class_mapping", "metric_args")
        for _key in check_keys:
            if _key not in data:
                raise ValueError(f"Missing {_key} in the data dictionary")

        # check the common set of images
        common_ids = set(list(data["prediction"].keys())).difference(
            set(list(data["ground_truth"].keys()))
        )

        if common_ids:
            raise ValueError(
                "prediction and ground truth does not have corresponding samples"
            )

    @staticmethod
    def _preprocess(data: dict, get_label=False, get_pred_scores=False) -> tuple:
        """
        A function to preprocess

        :param data: dictionary data
        :type data: dict
        :param get_label: boolean value to get the label
        :type get_label: bool
        :param get_pred_scores: boolean value to get the prediction scores
        :type get_pred_scores: bool

        :return: gt, pred
        :rtype: tuple(dict, dict)
        """
        gt_boxes, pred_boxes = {}, {}

        for image_id in data["ground_truth"]:
            for _ant in data["ground_truth"][image_id]["annotation"]:
                points = _ant["points"]
                box_points = [
                    points[0]["x"],
                    points[0]["y"],
                    points[1]["x"],
                    points[1]["y"],
                ]

                if get_label:
                    box_points.append(_ant["label"])

                if image_id in gt_boxes:
                    gt_boxes[image_id].append(box_points)
                else:
                    gt_boxes[image_id] = [box_points]

            for pred in data["prediction"][image_id]["objects"]:
                points = pred["box"]
                box_points = [points["x1"], points["y1"], points["x2"], points["y2"]]

                if get_label:
                    box_points.append(pred["prediction_class"])

                if get_pred_scores:
                    box_points.append(pred["confidence"])

                if image_id in pred_boxes:
                    pred_boxes[image_id].append(box_points)
                else:
                    pred_boxes[image_id] = [box_points]

        return (gt_boxes, pred_boxes)

    def _calc_precision_recall(self, gt_boxes, pred_boxes, threshold: float) -> tuple:
        """
        A function to calculate the precision and recall

        :param gt_boxes:
        :type gt_boxes:
        :param pred_boxes:
        :type pred_boxes:
        :param threshold:
        :type threshold:

        :return: calculated precision and recall
        :rtype: tuple
        """
        num_gt_boxes, num_pred_boxes = len(gt_boxes), len(pred_boxes)
        true_positives, false_positives = 0, 0

        for pred_box in pred_boxes:
            max_iou = 0
            for gt_box in gt_boxes:
                iou = self.iou.calculate(pred_box, gt_box)
                max_iou = max(max_iou, iou)

            if max_iou >= threshold:
                true_positives += 1
            else:
                false_positives += 1

        precision = (
            true_positives / (true_positives + false_positives)
            if (true_positives + false_positives) > 0
            else 0
        )
        recall = true_positives / num_gt_boxes

        return (precision, recall)

    def _calc_mAP_mAR(
        self, gt_boxes_dict: dict, pred_boxes_dict: dict, thresholds: list
    ) -> dict:
        """
        A function to calculate average precision and average recall

        :param gt_boxes_dict: image id and box points for ground truth
        :type gt_boxes_dict: dict
        :param pred_boxes_dict: image id and box points for prediction
        :type pred_boxes_dict: dict
        :param thresholds: list of threshold
        :type thresholds: list

        :return: dicitonary of results
        :rtype: dict
        """
        results = {"metric": [], "threshold": [], "value": []}
        for threshold in thresholds:
            image_precisions, image_recalls = [], []

            for image_id in gt_boxes_dict.keys():
                gt_boxes = gt_boxes_dict[image_id]
                pred_boxes = pred_boxes_dict[image_id]
                precision, recall = self._calc_precision_recall(
                    gt_boxes, pred_boxes, threshold
                )
                image_precisions.append(precision)
                image_recalls.append(recall)

            average_precision = np.mean(image_precisions)
            average_recall = np.mean(image_recalls)
            results["metric"].append("mAP")
            results["value"].append(average_precision)
            results["metric"].append("mAR")
            results["value"].append(average_recall)
            results["threshold"].append(threshold)
            results["threshold"].append(threshold)

        return pd.DataFrame(results)

    def __calculate_metrics(self, data: dict, class_mapping: dict) -> dict:
        """
        A function to  calculate the metrics

        :param data: data dicationry containing the data
        :type data: dict
        :param class_mapping: a dictionary with class mapping labels
        :type class_mapping: dict

        :return: results calculated
        :rtype: dict
        """
        results = {}

        # preprocess the data to convert the dictionasry into gt box
        # pred box
        gt_boxes, pred_boxes = self._preprocess(data)

        # get the threshold
        thresholds = data["metric_args"]["threshold"]

        if not isinstance(thresholds, list) and thresholds is not None:
            thresholds = [thresholds]

        # calculate the mAR
        results = self._calc_mAP_mAR(gt_boxes, pred_boxes, thresholds)

        return results

    def calculate(self, data: dict) -> dict:
        """
        Calculates the average precision score for the given dataset

        :param data: The input data required for calculation and plotting
                    {"prediction":, "ground_truth": , "class_mapping":}
        :type data: dict

        :return: calculated metric results
        :rtype: dict
        """
        result = {}

        # validate the data
        self._validate_data(data)

        # calculate results
        result = self.__calculate_metrics(data, data.get("class_mapping"))

        return {"result": result}


class PlotAvgPrecision:
    def __init__(self, data: dict, cohort_id: Optional[int] = None):
        self.data = data
        self.cohort_id = cohort_id

    def _validate_data(self):
        """
        validates the data required for plotting the bar plot
        """
        if not isinstance(self.data["result"], pd.DataFrame):
            raise ValueError(f"Data must be a data frame.")

    def save(self, fig: Figure, filename: str) -> str:
        """
        Saves the plot to a file

        :param fig: the matplotlib figure object where the plot image is stored
        :type fig: str

        :return: Path where the plot image is saved
        :rtype: str
        """
        dir_path = "plots"
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        if self.cohort_id:
            filepath = f"{dir_path}/{self.cohort_id}_{filename}"
        else:
            filepath = f"{dir_path}/{filename}"

        fig.savefig(filepath, format="png")
        return filepath

    def plot(self) -> Figure:
        """
        Plots the AUC curves.
        """
        sns.set_style("whitegrid")

        # validate the data
        self._validate_data()

        fig, ax = plt.subplots(figsize=(10, 7))
        sns.barplot(
            x="metric",
            y="value",
            hue="threshold",
            data=self.data["result"],
            ax=ax,
            palette="pastel",
        )
        ax.set_xlabel("Metric", fontdict={"fontsize": 14, "fontweight": "medium"})
        ax.set_ylabel("Value", fontdict={"fontsize": 14, "fontweight": "medium"})

        if self.cohort_id:
            title_str = f"mAP and mAR at Thresholds: cohort - {self.cohort_id}"
        else:
            title_str = "mAP and mAR at Thresholds"

        ax.set_title(
            title_str,
            fontdict={"fontsize": 16, "fontweight": "medium"},
        )
        ax.legend(loc="lower right")
        return fig


problem_type_map = {
    "classification": Classification,
    "semantic_segmentation": SemanticSegmentation,
    "object_detection": ObjectDetection,
}


@metric_manager.register("object_detection.average_precision")
def calculate_avg_precision(data: dict, problem_type: str):
    """
    A wrapper function to calculate the average precision
    """
    _metric_calculator = problem_type_map[problem_type]()
    result = _metric_calculator.calculate(data)
    return result


@plot_manager.register("object_detection.average_precision")
def plot_avg_precision(
    results: dict,
    save_plot: bool,
    file_name: str = "average_precision.png",
    cohort_id: Optional[int] = None,
) -> Union[str, None]:
    """
    A wrapper function to plot average precision

    :param results: Dictionary of the results
    :type results: dict
    :param save_plot: boolean value to save plot
    :type save_plot: bool

    :return: None or path to the saved plot
    :rtype: Union[str, None]
    """
    plotter = PlotAvgPrecision(data=results, cohort_id=cohort_id)
    fig = plotter.plot()
    if save_plot:
        return plotter.save(fig, filename=file_name)
    else:
        plt.show()
